// Copyright (c) 1999, A.H. van den Boogaard

#ifndef ELEMENTS_H
#define ELEMENTS_H

#include "material.h"
#include "geometry.h"
#include "shape.h"
#include "myelmbase.h"

#ifdef FT_GUI
#include "drawshape.h"
#endif

int check_mgd_ptrs( const FeaTure::Material_type *mat,
                    const FeaTure::Geometry *geom,
                    bool m, bool g );

class Truss:public FiniteElement
{
   private:
      FeaTure::Structural_Point *mat_point;

   protected:
      FeaTure::Material_type  *material_ptr;
      FeaTure::Geometry       *geometry_ptr;

      LocalDof* GetLocalDofs() const;
      bool NeedExpandStrain() const;

   public:
      Truss( int nr, Node_ptr *Nodes, FeaTure::Material_type *mat,
             FeaTure::Geometry *geom );
      ~Truss();

      static int reqd_nr_nodes() { return 2; }
      static int check_data( const FeaTure::Material_type *mat,
                             const FeaTure::Geometry *geom );
      FeaTure::Matrix GetLocalStiff( FeaTure::Vector& correction ) const;
      FeaTure::Matrix GetLocalMass( bool lumped ) const;
      FeaTure::Vector GetLocalLHS() const;
      bool symmetric() const { return material_ptr->symmetric(); }
      void iteration_update();
      void finish_increment();
      void WriteElementRestartData( std::ostream& restart ) const;
      void ReadElementRestartData( std::istream& restart );
      void PrintStressHeader( std::ostream& out ) const;
      void PrintXpStressHeader( std::ostream& out ) const;
      void PrintStress( std::ostream& out ) const;
      void PrintXpStress( std::ostream& out ) const;
#ifdef FT_GUI
      void Draw( wxDC& dc, const DeviceMap& map, bool numbered ) const;
      void DrawDispl( wxDC& dc, const DeviceMap& map, bool numbered,
                      float magnif ) const;
#endif
};

class LinBeam2d:public FiniteElement
{
   private:
      FeaTure::Vector LHS_begin;
      FeaTure::Vector LHS_end;

   protected:
      FeaTure::Material_type  *material_ptr;
      FeaTure::Geometry       *geometry_ptr;

      LocalDof* GetLocalDofs() const;
      FeaTure::Matrix get_N_mat( double L2 ) const;
      FeaTure::Matrix get_B_mat( double L2 ) const;
      FeaTure::Vector get_gen_forces( double L2 ) const;

   public:
      LinBeam2d( int nr, Node_ptr *Nodes, FeaTure::Material_type *mat,
                 FeaTure::Geometry *geom );
      ~LinBeam2d();

      static int reqd_nr_nodes() { return 2; }
      static int check_data( const FeaTure::Material_type *mat,
                             const FeaTure::Geometry *geom );
      FeaTure::Matrix GetLocalStiff( FeaTure::Vector& correction ) const;
      FeaTure::Vector GetLocalLHS() const;
      FeaTure::Matrix GetLocalMass( bool lumped ) const;
      bool symmetric() const { return material_ptr->symmetric(); }
      void iteration_update();
      void finish_increment();
      void WriteElementRestartData( std::ostream& restart ) const;
      void ReadElementRestartData( std::istream& restart );
      void PrintStressHeader( std::ostream& out ) const;
      void PrintXpStressHeader( std::ostream& out ) const;
      void PrintStress( std::ostream& out ) const;
      void PrintXpStress( std::ostream& out ) const;
#ifdef FT_GUI
      void Draw( wxDC& dc, const DeviceMap& map, bool numbered ) const;
      void DrawDispl( wxDC& dc, const DeviceMap& map, bool numbered,
                      float magnif ) const;
#endif
};

class Mass:public FiniteElement
{
   protected:
      Mass( int nr, Node_ptr *Nodes, FeaTure::Material_type *mat,
            FeaTure::Geometry *geom,
            int nr_dofs );
      FeaTure::Mass_type  *material_ptr;

   public:
      virtual ~Mass(){}

      static int reqd_nr_nodes() { return 1; }
      static int check_data( const FeaTure::Material_type *mat,
                             const FeaTure::Geometry *geom );
      FeaTure::Matrix GetLocalStiff( FeaTure::Vector& correction ) const;
      FeaTure::Vector GetLocalLHS() const;
      FeaTure::Matrix GetLocalMass( bool lumped ) const;
      bool symmetric() const { return true; }
      void iteration_update(){ return;}
      void finish_increment(){return;}
      void WriteElementRestartData( std::ostream& /*restart*/ ) const{}
      void ReadElementRestartData( std::istream& /*restart*/ ){}
      void PrintStressHeader( std::ostream& /*out*/ ) const {return;}
      void PrintXpStressHeader( std::ostream& /*out*/ ) const {return;}
      void PrintStress( std::ostream& /*out*/ ) const {return;}
      void PrintXpStress( std::ostream& /*out*/ ) const {return;}
#ifdef FT_GUI
      void Draw( wxDC& dc, const DeviceMap& map, bool numbered ) const;
      void DrawDispl( wxDC& dc, const DeviceMap& map, bool numbered,
                      float magnif ) const;
#endif
};

class Mass2D:public Mass
{
   protected:
      LocalDof* GetLocalDofs() const;

   public:
      Mass2D( int nr, Node_ptr *Nodes, FeaTure::Material_type *mat,
              FeaTure::Geometry *geom );
      ~Mass2D(){}
};

class Mass3D:public Mass
{
   protected:
      LocalDof* GetLocalDofs() const;

   public:
      Mass3D( int nr, Node_ptr *Nodes, FeaTure::Material_type *mat,
              FeaTure::Geometry *geom );
      ~Mass3D(){}
};

class NumIntElement:public FiniteElement
{
   protected:
      NumIntElement();
      NumIntElement( int nr, Node_ptr *Nodes, int nr_nodes, int nr_dofs,
                     FeaTure::Material_type *mat, FeaTure::Geometry *geom,
                     int nip );
      ~NumIntElement();

      FeaTure::Material_type  *material_ptr;
      FeaTure::Geometry       *geometry_ptr;

      FeaTure::Material_Point **mat_point;
      FeaTure::Shape          *shape;
#ifdef FT_GUI
      DrawShape      *draw_shape;
#endif

      FeaTure::IntLocs get_int_locs() const { return shape->get_int_locs(); }
      int get_nr_int_points() const { return shape->get_nr_int_points(); }
      void finish_increment();
      void WriteElementRestartData( std::ostream& restart ) const;
      void ReadElementRestartData( std::istream& restart );
      bool symmetric() const { return material_ptr->symmetric(); }
      void GetCoordinateMatrix( FeaTure::Matrix& coor, int ndim ) const;
      void GetNewCoordinateMatrix( FeaTure::Matrix& coor, int ndim ) const;
   public:
#ifdef FT_GUI
      void Draw( wxDC& dc, const DeviceMap& map, bool numbered ) const;
      void DrawDispl( wxDC& dc, const DeviceMap& map, bool numbered,
                      float magnif ) const;
#endif
};

class NumIntStructElement: virtual public NumIntElement
{
   protected:
      int nr_struct_dofs;
      FeaTure::StressType stype;

      NumIntStructElement( FeaTure::StressType st ):NumIntElement(),stype(st){}
      NumIntStructElement( int nr, Node_ptr *Nodes, int nr_nodes,
                           int struct_dofs, int nr_dofs,
                           FeaTure::Material_type *mat, FeaTure::Geometry *geom,
                           FeaTure::StressType st, int nip );
      ~NumIntStructElement();

      FeaTure::Matrix get_B_mat( const FeaTure::Vector &natcor,
                                 const FeaTure::Matrix &coor,
                                 double *detj ) const;
      FeaTure::Matrix get_N_mat( const FeaTure::Vector &natcor ) const;
      bool NeedExpandStrain() const;
      LocalDof* GetLocalDofs() const;
      void iteration_update_ip_struct( const FeaTure::Vector& natcor,
                                       const FeaTure::Vector& dU,
                                       FeaTure::Structural_Point *mat_point );

   public:
      FeaTure::Matrix GetLocalStiff( FeaTure::Vector& correction ) const;
      FeaTure::Vector GetLocalLHS() const;
      FeaTure::Matrix GetLocalMass( bool lumped ) const;
      void iteration_update();
      FeaTure::Vector GetLocalPresVec( int surface, double magnitude ) const;

      void PrintStress( std::ostream& out ) const;
      void PrintXpStress( std::ostream& out ) const;
      void PrintPostHeader( std::ostream &out, int stepnr ) const;
      void PrintPostFile( std::ostream& out ) const;
      void AddNodalPostContribution() const;

      static int check_data( const FeaTure::Material_type *mat,
                             const FeaTure::Geometry *geom,
                             FeaTure::StressType stype );
};

class NumIntThermalElement: virtual public NumIntElement
{
   protected:
      int nr_thermal_dofs;
      FeaTure::StressType dimtype;

      NumIntThermalElement( FeaTure::StressType st ):NumIntElement(),
                                                     dimtype(st){}
      NumIntThermalElement( int nr, Node_ptr *Nodes, int nr_nodes,
                            int nr_therm_dofs, int nr_dofs,
                            FeaTure::Material_type *mat,
                            FeaTure::Geometry *geom,
                            FeaTure::StressType st,
                            int nip );
      ~NumIntThermalElement();

      FeaTure::Matrix get_dNdX( const FeaTure::Vector &natcor,
                                const FeaTure::Matrix &coor,
                                double *detj ) const;
      LocalDof* GetLocalDofs() const;
      void iteration_update_ip_therm( const FeaTure::Vector& natcor,
                                      const FeaTure::Vector& dU,
                                      FeaTure::Thermal_Point *mat_point );

   public:
      FeaTure::Matrix GetLocalStiff( FeaTure::Vector& correction ) const;
      FeaTure::Vector GetLocalLHS() const;
      void iteration_update();
      void PrintStress( std::ostream& out ) const;
      void PrintXpStress( std::ostream& out ) const;

      void PrintStressHeader( std::ostream& out ) const;
      void PrintXpStressHeader( std::ostream& out ) const;
};

class NumIntThermoMechanicalElement:public NumIntStructElement,
                                           NumIntThermalElement
{
   protected:
      NumIntThermoMechanicalElement( int nr, Node_ptr *Nodes, int nr_nodes,
                                     int struct_dofs, int therm_dofs,
                                     FeaTure::Material_type *mat,
                                     FeaTure::Geometry *geom,
                                     FeaTure::StressType st,
                                     int nip );
      ~NumIntThermoMechanicalElement();

      LocalDof* GetLocalDofs() const;

   public:
      FeaTure::Matrix GetLocalStiff( FeaTure::Vector& correction ) const;
      FeaTure::Vector GetLocalLHS() const;
      void iteration_update();
      void PrintStress( std::ostream& out ) const;
      void PrintXpStress( std::ostream& out ) const;
};

class NumInt2D:public NumIntStructElement
{
   protected:
      NumInt2D( int nr, Node_ptr *Nodes, int nr_nodes, int nr_dofs,
                FeaTure::Material_type *mat, FeaTure::Geometry *geom,
                FeaTure::StressType st, int nip );
      ~NumInt2D();

   public:
      void PrintStressHeader( std::ostream& out ) const;
      void PrintXpStressHeader( std::ostream& out ) const;
      void PrintPostHeader( std::ostream &out, int stepnr ) const;
};

class NumInt3D:public NumIntStructElement
{
   protected:
      NumInt3D( int nr, Node_ptr *Nodes, int nr_nodes, int nr_dofs,
                FeaTure::Material_type *mat, FeaTure::Geometry *geom, int nip );
      ~NumInt3D();

   public:
      void PrintStressHeader( std::ostream& out ) const;
      void PrintXpStressHeader( std::ostream& out ) const;
      void PrintPostHeader( std::ostream &out, int stepnr ) const;
};

class NumIntThermMech2D:public NumIntThermoMechanicalElement
{
   protected:
      NumIntThermMech2D( int nr, Node_ptr *Nodes, int nr_nodes,
                         int struct_dofs, int therm_dofs,
                         FeaTure::Material_type *mat,
                         FeaTure::Geometry *geom, 
                         FeaTure::StressType st,
                         int nip );
      ~NumIntThermMech2D();

   public:
      void PrintStressHeader( std::ostream& out ) const;
      void PrintXpStressHeader( std::ostream& out ) const;
};

class Tri:public NumInt2D
{
   protected:
      Tri( int nr, Node_ptr *Nodes, int nr_nodes, int nr_dofs,
           FeaTure::Material_type *mat,
           FeaTure::Geometry *geom,
           FeaTure::StressType st, int nip );
      ~Tri(){}
};

class Tri3:public Tri
{
   protected:
      Tri3( int nr, Node_ptr *Nodes, FeaTure::Material_type *mat,
            FeaTure::Geometry *geom, FeaTure::StressType st, int nip );
      ~Tri3(){}

   public:
      static int reqd_nr_nodes() { return 3; }
};

class Tri6:public Tri
{
   protected:
      Tri6( int nr, Node_ptr *Nodes, FeaTure::Material_type *mat,
            FeaTure::Geometry *geom, FeaTure::StressType st, int nip, bool );
      ~Tri6(){}

   public:
      static int reqd_nr_nodes() { return 6; }
};

class Tri3ThermMech:public NumIntThermMech2D
{
   protected:
      Tri3ThermMech( int nr, Node_ptr *Nodes, FeaTure::Material_type *mat,
                     FeaTure::Geometry *geom, FeaTure::StressType st, int nip );
      ~Tri3ThermMech(){}

   public:
      static int reqd_nr_nodes() { return 3; }
};

class Tri3therm:public NumIntThermalElement
{
   public:
      Tri3therm( int nr, Node_ptr *Nodes, FeaTure::Material_type *mat,
                 FeaTure::Geometry *geom );
      ~Tri3therm(){}

      static int reqd_nr_nodes() { return 3; }
      static int check_data( const FeaTure::Material_type *mat,
                             const FeaTure::Geometry *geom );
};

class Quad:public NumInt2D
{
   protected:
      Quad( int nr, Node_ptr *Nodes, int nr_nodes, int nr_dofs,
            FeaTure::Material_type *mat, FeaTure::Geometry *geom,
            FeaTure::StressType st, int nxi, int neta );
      ~Quad(){}
};

class Quad4:public Quad
{
   protected:
      Quad4( int nr, Node_ptr *Nodes, FeaTure::Material_type *mat,
             FeaTure::Geometry *geom, FeaTure::StressType st,
             int nxi, int neta );
      ~Quad4(){}

   public:
      static int reqd_nr_nodes() { return 4; }
};

class Quad4ThermMech:public NumIntThermMech2D
{
   protected:
      Quad4ThermMech( int nr, Node_ptr *Nodes, FeaTure::Material_type *mat,
                      FeaTure::Geometry *geom, FeaTure::StressType st,
                      int nxi, int neta );
      ~Quad4ThermMech(){}

   public:
      static int reqd_nr_nodes() { return 4; }
};

class Quad8:public Quad
{
   protected:
      Quad8( int nr, Node_ptr *Nodes, FeaTure::Material_type *mat,
             FeaTure::Geometry *geom, FeaTure::StressType st,
             int nxi, int neta, bool );
      ~Quad8(){}

   public:
      static int reqd_nr_nodes() { return 8; }
};

class Tri3psn:public Tri3
{
   public:
      Tri3psn( int nr, Node_ptr *Nodes, FeaTure::Material_type *mat,
               FeaTure::Geometry *geom );
      ~Tri3psn(){}
};

class Tri3pss:public Tri3
{
   public:
      Tri3pss( int nr, Node_ptr *Nodes, FeaTure::Material_type *mat,
               FeaTure::Geometry *geom );
      ~Tri3pss(){}
};

class Tri3axi:public Tri3
{
   public:
      Tri3axi( int nr, Node_ptr *Nodes, FeaTure::Material_type *mat,
               FeaTure::Geometry *geom );
      ~Tri3axi(){}
};

class Tri6psn:public Tri6
{
   public:
      Tri6psn( int nr, Node_ptr *Nodes, FeaTure::Material_type *mat,
               FeaTure::Geometry *geom, bool );
      ~Tri6psn(){}
};

class Tri6pss:public Tri6
{
   public:
      Tri6pss( int nr, Node_ptr *Nodes, FeaTure::Material_type *mat,
               FeaTure::Geometry *geom, bool );
      ~Tri6pss(){}
};

class Tri6axi:public Tri6
{
   public:
      Tri6axi( int nr, Node_ptr *Nodes, FeaTure::Material_type *mat,
               FeaTure::Geometry *geom, bool );
      ~Tri6axi(){}
};

class Quad4pss:public Quad4
{
   public:
      Quad4pss( int nr, Node_ptr *Nodes, FeaTure::Material_type *mat,
                FeaTure::Geometry *geom );
      ~Quad4pss(){}
};

class Quad4psn:public Quad4
{
   public:
      Quad4psn( int nr, Node_ptr *Nodes, FeaTure::Material_type *mat,
                FeaTure::Geometry *geom );
      ~Quad4psn(){}
};

class Quad4axi:public Quad4
{
   public:
      Quad4axi( int nr, Node_ptr *Nodes, FeaTure::Material_type *mat,
                FeaTure::Geometry *geom );
      ~Quad4axi(){}
};

class Quad8pss:public Quad8
{
   public:
      Quad8pss( int nr, Node_ptr *Nodes, FeaTure::Material_type *mat,
                FeaTure::Geometry *geom, bool );
      ~Quad8pss(){}
};

class Quad8psn:public Quad8
{
   public:
      Quad8psn( int nr, Node_ptr *Nodes, FeaTure::Material_type *mat,
                FeaTure::Geometry *geom, bool );
      ~Quad8psn(){}
};

class Quad8axi:public Quad8
{
   public:
      Quad8axi( int nr, Node_ptr *Nodes, FeaTure::Material_type *mat,
                FeaTure::Geometry *geom, bool );
      ~Quad8axi(){}
};

class Quad4pssTherm:public Quad4ThermMech
{
   public:
      Quad4pssTherm( int nr, Node_ptr *Nodes, FeaTure::Material_type *mat,
                     FeaTure::Geometry *geom );
      ~Quad4pssTherm(){}

      static int reqd_nr_nodes() { return 4; }
};

class Quad4psnTherm:public Quad4ThermMech
{
   public:
      Quad4psnTherm( int nr, Node_ptr *Nodes, FeaTure::Material_type *mat,
                     FeaTure::Geometry *geom );
      ~Quad4psnTherm(){}
};

class Tri3psnTherm:public Tri3ThermMech
{
   public:
      Tri3psnTherm( int nr, Node_ptr *Nodes, FeaTure::Material_type *mat,
                    FeaTure::Geometry *geom );
      ~Tri3psnTherm(){}
};

class Hex:public NumInt3D
{
   public:
      Hex( int nr, Node_ptr *Nodes, int nr_nodes, int nr_dofs,
           FeaTure::Material_type *mat, FeaTure::Geometry *geom,
           int nxi, int neta, int nzeta );
      ~Hex(){}
};

class Hex8:public Hex
{
   public:
      Hex8( int nr, Node_ptr *Nodes, FeaTure::Material_type *mat,
            FeaTure::Geometry *geom );
      ~Hex8(){}

      static int reqd_nr_nodes() { return 8; }
};

#endif /* ELEMENTS_H */
