// Copyright (c) 1999, A.H. van den Boogaard

/*  
   Implementation of the LoadBar class. A LoadBar is a window in which
   a small size histogram can be plotted, e.g. to present the variation
   of certain data in time.
*/

// For compilers that support precompilation, includes "wx.h".
#include <wx/wxprec.h>
#ifdef __BORLANDC__
#pragma hdrstop
#endif
#ifndef WX_PRECOMP
#include <wx/wx.h>
#endif

#include "loadbar.h"
#include <cmath>
#include <algorithm>

IMPLEMENT_DYNAMIC_CLASS( LoadBar, wxWindow )

BEGIN_EVENT_TABLE( LoadBar, wxWindow )
  EVT_PAINT(  LoadBar::OnPaint )
END_EVENT_TABLE()

LoadBar::LoadBar(wxWindow *parent, const wxPoint& pos, const wxSize& size,
       	         long style, const char* name )
            : wxWindow( parent, -1, pos, size, style, name )
{
   GetClientSize( &width, &height );

   linewidth = 5;
   logscale = false;
   loads.reserve( (width/linewidth) );
   use_crit = false;
   crit     = 0.0;

   SetBackgroundColour(*wxWHITE);

   Reset();
}

LoadBar::~LoadBar()
{
}

void LoadBar::OnPaint( wxPaintEvent& WXUNUSED(event) )
{
   wxPaintDC dc( this );
   GetClientSize( &width, &height );

   wxPen *red_pen = wxThePenList->FindOrCreatePen("RED", linewidth, wxSOLID);
   wxPen *black_pen = wxThePenList->FindOrCreatePen("BLACK", linewidth, wxSOLID);

   red_pen->SetCap( wxCAP_BUTT );
   black_pen->SetCap( wxCAP_BUTT );

   wxCoord y0 = height;

   dc.Clear();

   for ( unsigned int i=0; i<loads.size(); i++ )
   {
      if ( loads[i].second )
         dc.SetPen(*red_pen);
      else
         dc.SetPen(*black_pen);

      wxCoord x, y;
      x = wxCoord((float(i)+0.5)*linewidth);
      y = y_screen( loads[i].first );

      dc.DrawLine( x, y0, x, y );
   }

   wxPen *blue_pen = wxThePenList->FindOrCreatePen("BLUE", 1, wxSOLID);
   dc.SetPen(*blue_pen);
   for ( int line=(int)ceil(graph_min); line<graph_max; line++ )
   {
      if ( line != 0 )
      {
         wxCoord y = y_screen( line );
         dc.DrawLine( 0, y, width, y );
      }
   }

   dc.SetPen(*wxCYAN_PEN);
   wxCoord y = y_screen( 0.0 );
   dc.DrawLine( 0, y, width, y );

   if ( use_crit )
   {
      dc.SetPen(*wxGREEN_PEN);
      wxCoord y = y_screen(crit);
      dc.DrawLine( 0, y, width, y );
   }
}

void LoadBar::Push( float value, bool highlight )
{
   if ( loads.size() >= loads.capacity() )
   {
      std::rotate( loads.begin(), loads.begin()+1, loads.end() );
      loads.pop_back();
      ScrollWindow(-linewidth,0,NULL);
   }

   if ( logscale )
   {
      float val;
      val = std::max( value, float(1.e-15) );
      value = log10(val);
   }

   loads.push_back( Value(value,highlight) );

   float maximum = (std::max_element( loads.begin(),loads.end() ))->first;
   float minimum = (std::min_element( loads.begin(),loads.end() ))->first;

   if ( use_crit )
   {
      maximum = std::max(maximum,crit);
      minimum = std::min(std::max(minimum,crit-2),crit);
   }

   maximum = ceil( std::max(maximum,float(1.0)) );
   minimum = floor( std::min(minimum,float(0.0)) );

   bool change_scale = ( graph_max != maximum || graph_min != minimum );

   graph_max = maximum;
   graph_min = minimum;
   factor = height/(graph_max-graph_min);

   wxClientDC dc( this );
   if ( change_scale ) Refresh();

   dc.BeginDrawing();
   wxPen *pen;

   wxCoord y0 = height;

   if ( highlight )
      pen = wxThePenList->FindOrCreatePen("RED", linewidth, wxSOLID);
   else
      pen = wxThePenList->FindOrCreatePen("BLACK", linewidth, wxSOLID);

   dc.SetPen(*pen);

   wxCoord x, y;
   int i = loads.size()-1;
   x = wxCoord((float(i)+0.5)*linewidth);
   y = y_screen( loads[i].first );
   dc.DrawLine( x, y0, x, y );

   wxPen *blue_pen = wxThePenList->FindOrCreatePen("BLUE", 1, wxSOLID);
   dc.SetPen(*blue_pen);
   for ( int line=(int)floor(graph_min); line<graph_max; line++ )
   {
      if ( line != 0 )
      {
         wxCoord y = y_screen(line);
         dc.DrawLine(0,y,width,y);
      }
   }

   dc.SetPen(*wxCYAN_PEN);
   y = y_screen( 0.0 );
   dc.DrawLine(0,y,width,y);

   if ( use_crit )
   {
      dc.SetPen(*wxGREEN_PEN);
      wxCoord y = y_screen(crit);
      dc.DrawLine(0,y,width,y);
   }

   dc.EndDrawing();
}

void LoadBar::Reset()
{
   graph_max = 1.0;
   graph_min = 0.0;
   loads.resize(0);
   factor = height/(graph_max-graph_min);
}

void LoadBar::SetCriterion( float c )
{
   use_crit=true;
   if ( logscale )
   {
      if ( c < 1.e-15 ) return;
      crit = log10(c);
   }
   else
      crit=c;
}

