// Copyright (c) 1999, A.H. van den Boogaard

#include "mycontrol.h"

#include <iomanip>
#include <cstdlib>

#include "control.h"
#include "domain.h"
#include "sysvec.h"
#include "constant.h"
#include "ftutils.h"

using std::setw;
using std::endl;
using std::div;
using std::min;
using std::sqrt;
using std::ostream;
using std::cout;
using std::cerr;
using std::ios;

// template definitions
// (putting control.cpp before using std::setw resulted in errors
//  on Borland C++ 5.5)
#include "control.cpp"
#include "domain.cpp"

using FeaTure::Vector;
using FeaTure::Matrix;
using FeaTure::SystemVector;
using FeaTure::SparseSysMat;
using FeaTure::SparseVector;
using FeaTure::CR_Matrix;
using FeaTure::IncrementalControl;
using FeaTure::IncrementalIterativeControl;

// explicit instantiation (needed at least for g++ and Borland C++ 5.5)
template float IncrementalControl<MyDomain>::get_fraction_executed( int );
// and another one for another g++ version (?)
template void IncrementalControl<MyDomain>::set_increments(int, double, int);


void LinearControl::initialize( MyDomain &model, ostream& /*out*/ )
{
   model.make_global_dofs();
}

bool LinearControl::continuate( MyDomain &model, ostream& out,
                                ostream& /*post*/, int /*steps*/ )
{
   SystemVector load_vector;
   SparseSysMat *Systemptr;
   Vector       rhs, solution;

   int loadset=1;
   int    nr_free;
   nr_free = model.get_nr_free();
   rhs.resize( nr_free );
   model.add_load_size( loadset, 1.0 );
   Systemptr = new SparseSysMat( nr_free, !(model.need_non_symmetric()) );
   model.init_system( *Systemptr );
   model.make_system( *Systemptr, rhs );
   model.get_load_vector( load_vector );

   rhs += load_vector.free();

   Systemptr->solve( rhs, solution );
   // MyDomain model is actually made for incremental analysis
   model.add_to_increments( solution );
   model.iteration_update();
   model.print_results( out );

   has_run = true;

   delete Systemptr;

   return false;  // no more steps to be done
}

EigenControl::~EigenControl()
{
   for ( size_t i=0; i<modes.size(); i++ ) delete modes[i];
}

void EigenControl::initialize( MyDomain &model, ostream& out )
{
   model.make_global_dofs();

   // calculate in "initialize" such that the different modes can be viewed
   // in the continuate function

   if ( model.need_non_symmetric() )
   {
      out << "Error: eigenfrequencies can only be determined for models with\n"
             "a symmetric stiffness matrix" << endl;
      cerr <<
           "Error: Cannot determine eigenfrequencies for a nonsymmetric matrix"
           << endl;
      throw EXIT_FAILURE;
   }

   SparseSysMat *K, *M;

   size_t    nr_free;
   nr_free = model.get_nr_free();
   K = new SparseSysMat( nr_free, true );
   M = new SparseSysMat( nr_free, true );
   Vector rhsdummy(nr_free,0);

   model.init_system( *K );
   model.make_system( *K, rhsdummy );
   model.make_mass_system( *M, false );  // consistent mass matrix

   /*
      shift:  K := K + lambdashift*M
      shift is in Hz and can be negative to converge to higher modes
      if shift is negative make also lambdashift negative
   */

   double lambdashift( PI2*std::abs(shift)*PI2*shift );
   if ( shift != 0.0 ) K->add(*M,lambdashift);

   size_t subspace_dim;
   subspace_dim = min(2*nr_frequencies,nr_frequencies+8);
   if ( nr_free < subspace_dim ) subspace_dim = nr_free;

   Vector lambda;
   Matrix V(nr_free,subspace_dim);

   subspace_iteration( *K, *M, &lambda, &V );

   // fill modes

   modes.reserve(subspace_dim);
   for ( size_t i=0; i<subspace_dim; i++ ) 
   {
      Mode *themode = new Mode(nr_free);
      themode->eigenvalue = lambda[i];
      for ( size_t j=0; j<nr_free; j++ )
         themode->mode[j] = V[j][i];
      modes.push_back( themode );
   }

   /*
      Eigenvalues are sorted in blocks; sort on absolute values to get the
      modes that are closest to minus the shift value (closest in lambda,
      not omega), then delete the excess modes above number of required
      frequencies.
   */

   std::sort( modes.begin(), modes.end(), CompareModePtrsAbsolute() );

   while ( modes.size() > nr_frequencies )
   {
      delete modes.back();
      modes.pop_back();
   }

   // now sort including the signed values
   std::sort( modes.begin(), modes.end(), CompareModePtrs() );

   // take account of the eigenvalue shift

   if ( shift != 0.0 )
   {
      std::vector<Mode*>::iterator themode;
      for (  themode = modes.begin(); themode != modes.end(); themode++ )
      {
         double thelambda = (*themode)->eigenvalue-lambdashift;
         // lambda=omega^2 take care that lambda>=0 also after roundoff
         if ( thelambda<0.0 && thelambda>-1.e-5*lambdashift ) thelambda = 0;
         (*themode)->eigenvalue = thelambda;
      }
   }

   delete K;
   delete M;

   has_run = true;
   current_mode = 0;
}

bool EigenControl::continuate( MyDomain &model, ostream& out,
                               ostream& /*post*/, int steps )
{
   if ( modes.size()==0 ) return false;

   if ( steps<0 ) steps = modes.size()-current_mode;

   for ( int i=0; i<steps; i++ )
   {
      if ( current_mode >= modes.size() ) current_mode = modes.size()-1;

      cout << "f(" << (current_mode+1) << ") = "
           << (sqrt(modes[current_mode]->eigenvalue)/PI2) << " Hz" << endl;
      out << "f(" << (current_mode+1) << ") = "
          << (sqrt(modes[current_mode]->eigenvalue)/PI2) << " Hz" << endl;

      // MyDomain model is actually made for incremental analysis
      model.reset_increments();
      model.add_to_increments( modes[current_mode]->mode );
      model.iteration_update();
      if ( ++current_mode >= modes.size() ) return false;
   }

   return ( current_mode < modes.size() );  // more steps to present
}

ExplicitControl::ExplicitControl():IncrementalControl<MyDomain>()
{
   DiagMassSystemptr = NULL;
}

ExplicitControl::~ExplicitControl()
{
   delete DiagMassSystemptr;
}

void ExplicitControl::initialize( MyDomain &model, ostream& out )
{
   model.make_global_dofs();

   int    nr_free;
   nr_free = model.get_nr_free();

   init_incremental( model );

   delta_u_1.resize( nr_free );
   delta_u_1.clear();

   DiagMassSystemptr = new FeaTure::DiagonalSysMat( nr_free );
   model.make_mass_system( *DiagMassSystemptr, true );  // lumped mass matrix

   if ( use_damping )
   {
      if ( Rayleigh_beta != 0.0 )
      {
         out << "In explicit analysis, damping can only be proportional to\n"
             << "the mass matrix, set Rayleigh coefficient Beta to zero" << endl;
         cerr << "error, see output file" << endl;
         throw EXIT_FAILURE;
      }
   }
}

void ExplicitControl::solve_increment( MyDomain &model, ostream& /*out*/, int loadset, double size )
{
   Vector solution;
   SystemVector lhs;

   model.add_load_size( loadset, size );
   model.make_LHS( lhs );

   SystemVector load;
   model.get_load_vector( load );
   rhs = (time_increment*time_increment)*(load.free() - lhs.free());

   DiagMassSystemptr->solve( rhs, solution );

   if ( use_damping ) delta_u_1 *= 1.0-0.5*time_increment*Rayleigh_alpha;
   solution += delta_u_1;
   if ( use_damping ) solution *= 1.0/(1.0+0.5*time_increment*Rayleigh_alpha);
   delta_u_1 = solution;

   model.add_to_increments( solution );
   model.iteration_update();
}

void ExplicitControl::log_increment_nr( ostream& /*out*/ )
{
   if ( div(incr_nr,100).rem == 0 ) cout << "increment " << setw(6) << incr_nr << endl;
}

                             
MyIncrementalIterativeControl::MyIncrementalIterativeControl()
                               :IncrementalIterativeControl<MyDomain>(){}
MyIncrementalIterativeControl::~MyIncrementalIterativeControl(){}

void MyIncrementalIterativeControl::print_norms( std::ostream& out )
{
   double ratio;
   if ( load_norm > 1E-12  ) ratio = unbalance_norm / load_norm;
   else
      if ( unbalance_norm < 1E-12 ) ratio = 0;
      else ratio = 1;

   out << "  unbalance norm: " << std::setw(11) << unbalance_norm
       << "  load norm: " << std::setw(11) << load_norm
       << "  ratio: " << std::setw(11) << ratio << std::endl;

#ifdef FT_GUI
   static int previous_increment_number=0;

   if ( nr_iter == 1 ) show_criterion( EPS_UNB );
   show_convergence( ratio,
                     nr_iter==1 && this->incr_nr!=previous_increment_number );

   previous_increment_number = this->incr_nr;
#endif
}


NonLinearStaticControl::NonLinearStaticControl():MyIncrementalIterativeControl()
{
   iteration_method = FULL_NEWTON_RAPHSON;
   Systemptr = NULL;
}

NonLinearStaticControl::~NonLinearStaticControl()
{
   delete Systemptr;
}

void NonLinearStaticControl::initialize( MyDomain &model, ostream& /*out*/ )
{
   model.make_global_dofs();
   init_incremental( model );

   /* Set up element dofs and initialize sparse system matrix */
   Systemptr = new SparseSysMat( model.get_nr_free(), !(model.need_non_symmetric()) );
   model.init_system( *Systemptr );
}

void NonLinearStaticControl::solve_increment( MyDomain &model, ostream& out, int loadset, double size )
{
   SystemVector Loads, lhs;

   nr_iter=0;

   model.add_load_size( loadset, size );

   model.make_system( *Systemptr, rhs );

   model.make_LHS( lhs );
   model.get_load_vector( Loads );

   rhs += Loads.free() - lhs.free();

   load_norm = rhs.norm();

   cout << "iteration ";

   try
   {
      do
      {
         nr_iter++;

         if ( !div( nr_iter-1,10).rem && nr_iter != 1 )
                                   cout << endl << "          " ;
         cout << " " << setw(3) << nr_iter;
         cout.flush();

         if ( nr_iter > 1 && iteration_method == FULL_NEWTON_RAPHSON )
            model.update_system( *Systemptr, rhs );

         solve_iteration( *Systemptr, model );
         model.make_LHS( lhs );
         model.get_load_vector( Loads );
         rhs = Loads.free() - lhs.free();

         calculate_norms();

         out.precision(4);
         out.setf( ios::scientific, ios::floatfield );
         out << endl << "iteration " << nr_iter << endl;
         print_norms(out);

#ifdef FT_GUI
         wxYield();
#endif

      } while ( !converged() && nr_iter < max_iter );
   }
   catch ( const FeaTure::divergence_error& e )
   {
      model.add_load_size( loadset, -size );
      model.reset_increments();
      model.iteration_update();
      throw;
   }

   if ( !converged() )
   {
      model.add_load_size( loadset, -size );
      model.reset_increments();
      model.iteration_update();
      throw( FeaTure::not_converged_error( "solution not converged" ) );;
   }

   cout << endl;
}

NonLinearNewmarkControl::NonLinearNewmarkControl()
   :MyIncrementalIterativeControl()
{
   iteration_method = FULL_NEWTON_RAPHSON;

   Newmark_alpha = 0.25;
   Newmark_delta = 0.5;

   Systemptr = NULL;
   MassSystemptr = NULL;
   DampSystemptr = NULL;
}

NonLinearNewmarkControl::~NonLinearNewmarkControl()
{
   delete Systemptr;
   delete MassSystemptr;
   delete DampSystemptr;
}

void NonLinearNewmarkControl::set_Newmark( double alpha, double delta )
{
   Newmark_alpha = alpha;
   Newmark_delta = delta;
}

void NonLinearNewmarkControl::initialize( MyDomain &model, ostream& /*out*/ )
{
   model.make_global_dofs();

   int    nr_free;
   nr_free = model.get_nr_free();

   init_incremental( model );

   velocity.resize( nr_free );
   velocity.clear();

   acceleration.resize( nr_free );
   acceleration.clear();

   /* Set up element dofs and initialize sparse system matrix */
   Systemptr = new SparseSysMat( nr_free, !(model.need_non_symmetric()) );
   model.init_system( *Systemptr );

   MassSystemptr = new SparseSysMat( nr_free, true );
   model.make_mass_system( *MassSystemptr, false ); // consistent mass matrix

   if ( use_damping )
      DampSystemptr = new SparseSysMat( nr_free, !(model.need_non_symmetric()) );
}

void NonLinearNewmarkControl::solve_increment( MyDomain &model, ostream& out, int loadset, double size )
{
   double delta, alpha, a0, a1, a2, a3, a4, a5, a6, a7;
   Vector incr_solution_free;
   SystemVector lhs, Loads;

   nr_iter=0;
   delta = Newmark_delta;
   alpha = Newmark_alpha;
   a2 = 1.0/(alpha*time_increment);
   a0 = a2/time_increment;
   a1 = a2*delta;
   a3 = 1.0/(2*alpha) - 1.0;
   a4 = delta/alpha - 1.0;
   a5 = 0.5*time_increment*(a4-1.0);
   a7 = delta*time_increment;
   a6 = time_increment-a7;

   model.add_load_size( loadset, size );
   model.get_load_vector( Loads );
   model.make_system( *Systemptr, rhs );

   if ( use_damping )
   {
      DampSystemptr->clear();
      if ( Rayleigh_alpha != 0.0 ) DampSystemptr->add( *MassSystemptr, Rayleigh_alpha );
      if ( Rayleigh_beta  != 0.0 ) DampSystemptr->add( *Systemptr, Rayleigh_beta );

      Systemptr->add( *DampSystemptr, a1 );
   }

   Systemptr->add( *MassSystemptr, a0 );

   model.make_LHS( lhs );

   rhs += Loads.free() - lhs.free() -
          (*MassSystemptr)*( -a2*velocity - a3*acceleration );

   if ( use_damping )
      rhs -= (*DampSystemptr)*( -a4*velocity - a5*acceleration );

   load_norm = rhs.norm();

   cout << "iteration ";

   do
   {
      nr_iter++;
      if ( nr_iter > 1 && iteration_method == FULL_NEWTON_RAPHSON )
      {
         model.update_system( *Systemptr, rhs );
         if ( use_damping )
         {
            DampSystemptr->clear();
            if ( Rayleigh_alpha != 0.0 ) DampSystemptr->add( *MassSystemptr, Rayleigh_alpha );
            if ( Rayleigh_beta  != 0.0 ) DampSystemptr->add( *Systemptr, Rayleigh_beta );

            Systemptr->add( *DampSystemptr, a1 );
         }
         Systemptr->add( *MassSystemptr, a0 );

#ifdef FT_GUI
         wxYield();
#endif
      }
      solve_iteration( *Systemptr, model );
      model.get_incremental_solution_vector( incr_solution_free );

      model.make_LHS( lhs );

      model.get_load_vector( Loads );
      rhs = Loads.free() - lhs.free() -
            (*MassSystemptr)*( a0*incr_solution_free - a2*velocity - a3*acceleration );

      if ( use_damping )
         rhs -= (*DampSystemptr)*( a1*incr_solution_free - a4*velocity - a5*acceleration );

      calculate_norms();

      if ( !div( nr_iter-1,10).rem && nr_iter != 1 )
                                   cout << endl << "          " ;
      cout << " " << setw(3) << nr_iter;
      cout.flush();
      out.precision(4);
      out.setf( ios::scientific, ios::floatfield );
      out << endl << "iteration " << nr_iter << endl;
      print_norms(out);

   } while ( !converged() && nr_iter < max_iter );

   Vector acc0(acceleration);

   acceleration = a0*incr_solution_free - a2*velocity - a3*acceleration;
   velocity += a6*acc0 + a7*acceleration;

   cout << endl;
}
