// Copyright (c) 1999, A.H. van den Boogaard

#include "mydomain.h"

#include <iomanip>

using std::ostream;
using std::setw;
using std::endl;

// template definitions
// (putting control.cpp before using std::setw resulted in errors
//  on Borland C++ 5.5)
#include "domain.cpp"

using FeaTure::SystemVector;
using FeaTure::Vector;
using FeaTure::PostProc;

MyDomain::MyDomain(){}

MyDomain::~MyDomain(){}

void MyDomain::print_model( ostream& out ) const
{
   print_title( out );
   print_nodes( out );
   print_elements( out );
   print_suppress( out );
   print_loads( out );
}

void MyDomain::print_title( ostream& out ) const
{
    out << endl << "TITLE" << endl << title << endl << endl;
}

void MyDomain::print_nodes( ostream& out ) const
{
   out << endl << "NODES" << endl;
   for ( ConstNodeIterator node=nodes.begin(); node!=nodes.end(); node++ )
      (*node)->print_clean( out );
}

void MyDomain::print_elements( ostream& out ) const
{
   out << endl << "ELEMENTS" << endl;

   for ( ConstElmIterator elm=elements.begin(); elm!=elements.end(); elm++ )
      (*elm)->print( out );
}

void MyDomain::print_suppress( ostream& out ) const
{
   out << endl << "SUPPRESS" << endl;
   for ( ConstNodeIterator node=nodes.begin(); node!=nodes.end(); node++ )
      (*node)->print_fixed_dofs( out );
}

void MyDomain::print_loads( ostream& out ) const
{
   out << endl << "LOADS" << endl;
   for ( ConstBoundCondIterator bc=loads.begin(); bc!=loads.end(); bc++ )
      (*bc)->print( out );
}

void MyDomain::print_results( std::ostream& out ) const
{
   print_solution( out );
   print_xp_stresses( out );
   print_reactions( out );
}

void MyDomain::print_solution( ostream& out ) const
{
   out << endl << "nodal solution in global coordinates" << endl << endl;

   MyNode::print_header( out );
   for ( ConstNodeIterator node=nodes.begin(); node!=nodes.end(); node++ )
      (*node)->print_solution( out );
}

void MyDomain::print_node_table( ostream& out ) const
{
   for ( ConstNodeIterator node=nodes.begin(); node!=nodes.end(); node++ )
      (*node)->print_table( out );
}

void MyDomain::print_stresses( ostream& out ) const
{
   out << endl << "element stresses" << endl;
   for ( ConstElmIterator elm=elements.begin(); elm!=elements.end(); elm++ )
   {
      (*elm)->PrintStressHeader( out );
      (*elm)->PrintStress( out );
   }
}

void MyDomain::print_xp_stresses( ostream& out ) const
{
   out << endl << "element stresses" << endl;
   for ( ConstElmIterator elm=elements.begin(); elm!=elements.end(); elm++ )
   {
      (*elm)->PrintXpStressHeader( out );
      (*elm)->PrintXpStress( out );
   }
}

void MyDomain::print_reactions( ostream& out ) const
{
   out << endl << "reactions" << endl;

   SystemVector Lhs, Loads;

   make_LHS( Lhs );
   get_load_vector( Loads );
   Lhs -= Loads;

   for ( ConstNodeIterator node=nodes.begin(); node!=nodes.end(); node++ )
      (*node)->print_reactions( out, Lhs.fixed(), Lhs.prescribed() );
}

void MyDomain::post_file( ostream& out, int stepnr, const PostProc &postproc ) const
{
   out.setf(std::ios::scientific,std::ios::floatfield);
   out.precision(3);

/*   FiniteElement::SetPostItem( FiniteElement::coordinates );
   (*elements.begin())->PrintPostHeader( out );
   for ( ConstElmIterator elm=elements.begin(); elm!=elements.end(); elm++ )
   {
      (*elm)->PrintPostFile( out );
   } */

   if ( postproc.need_displa() )
   {
      out << "displacements" << ' ' << setw(7) << 1 << ' ' << setw(11) << stepnr << ' ' <<
                setw(7) << 2 << ' ' << setw(7) << 1 << ' ' << setw(7)  << 1 << endl;
      out << "UX" << endl << "UY" << endl << "UZ" << endl;

      for ( ConstNodeIterator node=nodes.begin(); node!=nodes.end(); node++ )
      {
         out << setw(7) << (*node)->get_nr() <<
               setw(11) << (*node)->get_total_displacement() << endl;
      }
   }

   if ( postproc.need_incdis() )
   {
      out << "incr_displa" << ' ' << setw(7) << 1 << ' ' << setw(11) << stepnr << ' ' <<
                setw(7) << 2 << ' ' << setw(7) << 1 << ' ' << setw(7)  << 1 << endl;
      out << "Delta_UX" << endl << "Delta_UY" << endl << "Delta_UZ" << endl;

      for ( ConstNodeIterator node=nodes.begin(); node!=nodes.end(); node++ )
      {
         out << setw(7) << (*node)->get_nr() <<
               setw(11) << (*node)->get_incremental_displacement() << endl;
      }
   }

   if ( postproc.need_temp() )
   {
      out << "temperature" << ' ' << setw(7) << 1 << ' ' << setw(11) << stepnr << ' ' <<
                setw(7) << 1 << ' ' << setw(7) << 1 << ' ' << setw(7)  << 1 << endl;
      out << "Temp" << endl;

      for ( ConstNodeIterator node=nodes.begin(); node!=nodes.end(); node++ )
      {
         out << setw(7) << (*node)->get_nr() <<
               setw(11) << (*node)->get_temperature() << endl;
      }
   }

   if ( postproc.need_inctemp() )
   {
      out << "temp_increment" << ' ' << setw(7) << 1 << ' ' << setw(11) << stepnr << ' ' <<
                setw(7) << 1 << ' ' << setw(7) << 1 << ' ' << setw(7)  << 1 << endl;
      out << "Delta_Temp" << endl;

      for ( ConstNodeIterator node=nodes.begin(); node!=nodes.end(); node++ )
      {
         out << setw(7) << (*node)->get_nr() <<
               setw(11) << (*node)->get_temperature_increment() << endl;
      }
   }

   if ( postproc.need_stress() )
   {
      FiniteElement::SetPostItem( PostProc::stresses );
      for ( ConstNodeIterator node=nodes.begin(); node!=nodes.end(); node++ )
      {
         (*node)->dataVector.resize(1);  // will be resized by the next calls
         (*node)->dataVector.clear();
         (*node)->dataContribution = 0;
      }
      (*elements.begin())->PrintPostHeader( out, stepnr );
      for ( ConstElmIterator elm=elements.begin(); elm!=elements.end(); elm++ )
         (*elm)->AddNodalPostContribution();

      for ( ConstNodeIterator node=nodes.begin(); node!=nodes.end(); node++ )
      {
         /* GiD wants Sxy at the third place for 2D results */
         if ( (*node)->dataVector.size() == 4 )
         {
            double swap( (*node)->dataVector[3] );
            (*node)->dataVector[3] = (*node)->dataVector[2];
            (*node)->dataVector[2] = swap;
         }

         (*node)->dataVector *= 1.0/double((*node)->dataContribution);
         out << setw(7) << (*node)->get_nr() << setw(11);

         (*node)->dataVector.print_row( out );
         out << endl;
      }
   }

   if ( postproc.need_epspeq() )
   {
      FiniteElement::SetPostItem( PostProc::eps_plast_eq );
      for ( ConstNodeIterator node=nodes.begin(); node!=nodes.end(); node++ )
      {
         (*node)->dataVector.resize(1);
         (*node)->dataVector.clear();
         (*node)->dataContribution = 0;
      }
      (*elements.begin())->PrintPostHeader( out, stepnr );
      for ( ConstElmIterator elm=elements.begin(); elm!=elements.end(); elm++ )
      {
         (*elm)->AddNodalPostContribution();
      }

      for ( ConstNodeIterator node=nodes.begin(); node!=nodes.end(); node++ )
      {
         (*node)->dataVector *= 1.0/double((*node)->dataContribution);
         out << setw(7) << (*node)->get_nr() << setw(11);

         (*node)->dataVector.print_row( out );
         out << endl;
      }
   }
}

void MyDomain::get_extremes( double &minx, double &miny, double &minz,
                           double &maxx, double &maxy, double &maxz, float dscale ) const
{
   double x, y, z;

   for ( ConstNodeIterator node=nodes.begin(); node!=nodes.end(); node++ )
   {
      if ( dscale == 0.0 )
         (*node)->get_location( x, y, z );
      else
         (*node)->draw_location( x, y, z, dscale );
            
      if ( node==nodes.begin() )
      {
         minx = maxx = x;
         miny = maxy = y;
         minz = maxz = z;
      }
      else
      {
         minx = std::min( minx, x ); maxx = std::max( maxx, x );
         miny = std::min( miny, y ); maxy = std::max( maxy, y );
         minz = std::min( minz, z ); maxz = std::max( maxz, z );
      }
   }
}

#ifdef FT_GUI

void MyDomain::draw_displaced( wxDC& dc, const DeviceMap& map,
                               const DrawProperties &drawprops ) const
{
   double x, y, z;

   dc.SetFont( drawprops.elementFont );
   dc.SetTextForeground( drawprops.elementFontColor );

   for ( ConstElmIterator elm=elements.begin(); elm!=elements.end(); elm++ )
      (*elm)->DrawDispl( dc, map, drawprops.plot_element_nrs(), drawprops.displa_scale() );

   dc.SetFont( drawprops.nodeFont );
   dc.SetTextForeground( drawprops.nodeFontColor );

   for ( ConstNodeIterator node=nodes.begin(); node!=nodes.end(); node++ )
   {
      (*node)->draw_location( x, y, z, drawprops.displa_scale() );
      dc.DrawEllipse( map.XtoScreen(x)-1, map.YtoScreen(y)-1, 2, 2 );

      if ( drawprops.plot_node_nrs() )
      {
         char s[20];
         sprintf(s,"%d",(*node)->get_nr());
         dc.DrawText( s, map.XtoScreen(x), map.YtoScreen(y) );
      }
   }
}

void MyDomain::draw_nodes( wxDC& dc, const DeviceMap& map, bool numbered ) const
{
   double x, y, z;

   for ( ConstNodeIterator node=nodes.begin(); node!=nodes.end(); node++ )
   {
      (*node)->get_location( x, y, z );
      dc.DrawEllipse( map.XtoScreen(x)-1, map.YtoScreen(y)-1, 2, 2 );

      if ( numbered )
      {
         char s[20];
         sprintf(s,"%d",(*node)->get_nr());
         dc.DrawText( s, map.XtoScreen(x), map.YtoScreen(y) );
      }
   }
}

void MyDomain::draw_elements( wxDC& dc, const DeviceMap& map, bool numbered ) const
{
   for ( ConstElmIterator elm=elements.begin(); elm!=elements.end(); elm++ )
   {
      (*elm)->Draw( dc, map, numbered );
   }
}

#endif // FT_GUI
