// Copyright (c) 1999, A.H. van den Boogaard

#include "mynodes.h"

#include <iomanip>

using FeaTure::Vector;
using FeaTure::Dof;

using std::ostream;
using std::ios;
using std::cout;
using std::setw;
using std::endl;

void MyNode::print_clean( ostream& out ) const
{
   out.setf( ios::scientific, ios::floatfield );
   out.precision(4);

   out << setw(8) << get_nr() << "   " << setw(12) << get_location() << endl;
}

void MyNode::print() const
{
   int idof;

   cout.setf( ios::scientific, ios::floatfield );
   cout.precision(4);

   cout << "node nr. " << setw(3) << get_nr() << " : coordinates"
        << setw(12) << get_location() << endl;

   for ( idof=0; idof<nr_nodal_dofs; idof++ )
   {
      cout << "    dof nr. " << setw(3) << idof;
      dofs[idof]->print();
   }
   cout << endl;
}

void MyNode::print_fixed_dofs( ostream& out ) const
{
   int idof;

   for ( idof=0; idof<nr_nodal_dofs; idof++ )
   {
      if ( dofs[idof]->is_fixed() )
      {
         out << "  node " << setw(8) << get_nr() << "   ";
         Dof::Type type( dofs[idof]->get_dof_type() );
         print_dof_type( type, out );
         if ( dofs[idof]->is_vector() )
         {
            out.setf( ios::scientific, ios::floatfield );
            out.precision(4);
            out << setw(7) << dofs[idof]->get_direction();
         }
         out << endl;
      }
   }
}

void MyNode::print_solution( ostream& out ) const
{
   FeaTure::CartVector U(0,0,0), R(0,0,0), dir;
   double     value, T(0);
   bool       any_rotation = false, any_temp = false;

   for ( int idof=0; idof<nr_nodal_dofs; idof++ )
   {
      value = dofs[idof]->get_total();

      if ( dofs[idof]->get_dof_type() == Dof::u )
      {
         dir = dofs[idof]->get_direction();
         U += dir*value;
      }
      else if ( dofs[idof]->get_dof_type() == Dof::r )
      {
         any_rotation = true;
         dir = dofs[idof]->get_direction();
         R += dir*value;
      }
      else if ( dofs[idof]->get_dof_type() == Dof::t )
      {
         any_temp = true;
         T = value;
      }
   }

   out.setf( ios::scientific, ios::floatfield );
   out.precision(3);
   out << setw(7)  << get_nr() << " " << setw(11) << U;

   if ( any_rotation ) out << " " << setw(11) << R;
   else if ( any_temp ) out << "                                    ";

   if ( any_temp ) out << " " << setw(11) << T;

   out << endl;

   return;
}

void MyNode::print_table( ostream& out ) const
{
   double     value, T(0);
   bool       any_temp(false);

   for ( int idof=0; idof<nr_nodal_dofs; idof++ )
   {
      value = dofs[idof]->get_total();

      if ( dofs[idof]->get_dof_type() == Dof::t )
      {
         any_temp = true;
         T = value;
      }
   }

   out.setf( ios::scientific, ios::floatfield );
   out.precision(3);

   double x,y,z;
   get_location(x,y,z);
   out << setw(7) <<  y;

   if ( any_temp ) out << " " << setw(11) << T << endl;

   return;
}

void MyNode::print_header( ostream& out )
{
   out << "   node      U_X         U_Y         U_Z    ";
   out <<         "     R_X         R_Y         R_Z          T" << endl << endl;

   return;
}

void MyNode::print_reactions( ostream& out, const Vector& lhs_fixed,
                                          const Vector& lhs_prescribed ) const
{
   for ( int idof=0; idof<nr_nodal_dofs; idof++ )
   {
      out.precision(4);
      if ( dofs[idof]->is_fixed() )
      {
         out << setw(8) << get_nr() << "  ";
         print_dof_type( dofs[idof]->get_dof_type(), out );

         out << "  " << setw(11) << lhs_fixed(dofs[idof]->get_global_dof());
         if ( dofs[idof]->is_vector() )
         {
            out.setf( ios::fixed );
            out << "   " << setw(4) << dofs[idof]->get_direction();
            out.setf( ios::scientific, ios::floatfield );
         }
         out << endl;
      }

      if ( dofs[idof]->is_prescribed() )
      {
         out << setw(8) << get_nr() << "  ";
         print_dof_type( dofs[idof]->get_dof_type(), out );

         out << "  " << setw(11) << lhs_prescribed(dofs[idof]->get_global_dof());
         if ( dofs[idof]->is_vector() )
         {
            out.setf( ios::fixed );
            out << "   " << setw(4) << dofs[idof]->get_direction();
            out.setf( ios::scientific, ios::floatfield );
         }
         out << endl;
      }
   }
}


