// Copyright (c) 1999, A.H. van den Boogaard

#include "plastic.h"
#include <cmath>
#include <cassert>

using FeaTure::Matrix;
using FeaTure::Vector;
using FeaTure::CollectArray;
using FeaTure::StressType;
using FeaTure::gen;
using FeaTure::pss;
using FeaTure::psn;
using FeaTure::uni;
using FeaTure::axi;

using std::cerr;
using std::endl;
using std::sqrt;
using std::pow;

inline double square( double d ){ return d*d; }

VonMises_Point::VonMises_Point( VonMises_type *mat, StressType st )
                               :FeaTure::Structural_Point(st),parameters(mat)
{
   int xstress_size = ExpandedStressSize( st );

   e_elas0.resize( xstress_size );
   e_elas0.clear();
   e_elas1.resize( xstress_size );
   e_elas1.clear();

   strain_bar.resize( strain_bar_size() );
   strain_bar.clear();
   prev_dstrain.resize( StressSize( st ) );
   prev_dstrain.clear();

   eq0 = eq1 = e_tilde = p0 = p1 = 0;
   plastic = false;
}

VonMises_Point::~VonMises_Point(){}

Matrix VonMises_Point::get_D_mat() const
{
   return condense( get_xp_D_mat() );
}

Matrix VonMises_Point::get_xp_D_mat() const
{
   double E, nu, ep0, C, n, G, K, q, Q, R, H, B, KmQ3;
   int    xstrsize, i, j;

   E = parameters->get_Young();
   nu = parameters->get_Poisson();
   xstrsize = ExpandedStressSize( stress_type );

   if ( !plastic ) return iso_lin_elas_D_mat( E, nu, xstrsize, stress_type );

   ep0 = parameters->get_eps0();
   C = parameters->get_nadai_C();
   n = parameters->get_nadai_n();

   G = E/(2*(1+nu));
   K = E/(3*(1-2*nu));

   Vector S(2*G*e_elas1);
   for ( i=4; i<=xstrsize; i++ ) S(i) *= 0.5;

   q = S(1)*S(1) + S(2)*S(2) + S(3)*S(3);
   for ( i=4; i<=xstrsize; i++ ) q += 2*S(i)*S(i);
   q = sqrt( 1.5*q );
   H = C*n*pow(eq1+ep0,n-1);
   B = H/(3*G);
   Q = q/e_tilde;
   R = 1.5*(1-(eq1-eq0)*H/q)/(q*e_tilde*(1+B));
   KmQ3 = K-Q/3;

   Matrix D(xstrsize,xstrsize);

   for ( i=1; i<=xstrsize; i++ )
      for ( j=1; j<=xstrsize; j++ )
         D(i,j) = -R*S(i)*S(j);

   for ( i=1; i<=3; i++ )
   {
      D(i,i) += Q;
      for ( j=1; j<=3; j++ )
         D(i,j) += KmQ3;
   }

   for ( i=4; i<=xstrsize; i++ )
      D(i,i) += 0.5*Q;

   return D;
}

Vector VonMises_Point::get_stress() const
{
   return condense_vector( get_xp_stress() );
}

Vector VonMises_Point::get_xp_stress() const
{
   double E, nu, G;
   int    xstrsize, istr;

   E = parameters->get_Young();
   nu = parameters->get_Poisson();
   G = E/(2*(1+nu));
   Vector sigma( 2*G*e_elas1 );
   xstrsize = ExpandedStressSize( stress_type );
   for ( istr=4; istr<=xstrsize; istr++ ) sigma(istr) *= 0.5;
   for ( istr=1; istr<=3; istr++ ) sigma(istr) -= p1;
   return sigma;
}

void VonMises_Point::set_strain_increment( const Vector& strain_increment )
{
   int    i;
   double C, n, ep0, E, nu, G, K, sig_0, q, D_eps_p_eq, Df,
          Deps_V, Deps_V_3, H;
   int    maxiter = 10;
   double EPS_P   = 1E-5;

   Vector dstrain( strain_increment );

   Vector ddstrain( strain_increment - prev_dstrain );
   prev_dstrain = strain_increment;
   ExpandEps( dstrain, ddstrain, strain_bar );

   int xstrsize = ExpandedStressSize( stress_type );

   E = parameters->get_Young();
   nu = parameters->get_Poisson();
   ep0 = parameters->get_eps0();
   C = parameters->get_nadai_C();
   n = parameters->get_nadai_n();
   G = E/(2*(1+nu));

   Vector Delta_e( dstrain );

   Deps_V = Delta_e(1)+Delta_e(2)+Delta_e(3);
   Deps_V_3 = Deps_V/3;
   Delta_e(1) -= Deps_V_3; Delta_e(2) -= Deps_V_3; Delta_e(3) -= Deps_V_3;

   e_elas1 = e_elas0 + Delta_e;

   Vector S(2*G*e_elas1);
   for ( i=4; i<=xstrsize; i++ ) S(i) *= 0.5;

   q = S(1)*S(1) + S(2)*S(2) + S(3)*S(3);
   for ( i=4; i<=xstrsize; i++ ) q += 2*S(i)*S(i);
   q = sqrt( 1.5*q );
   e_tilde = 0.5*q/G;

   D_eps_p_eq = 0;
   eq1 = eq0;
   sig_0 = C*pow(eq0+ep0,n);

   plastic = ( q > sig_0 );

   if ( plastic )
   {
      int iter = 0;

      while ( std::fabs( Df=( G*(2*e_tilde-3*D_eps_p_eq)-sig_0 ) ) > EPS_P
              && iter++ < maxiter )
      {
         H = C*n*pow(eq1+ep0,n-1);
         D_eps_p_eq += Df/(3*G+H);
         eq1 = eq0 + D_eps_p_eq;
         sig_0 = C*pow(eq1+ep0,n);
      }
      q = sig_0;
      S *= 1/(1+3*G/q*D_eps_p_eq);

      /* e_elas1 = e_elas0 + D_e - D_e_plas */

      double alpha = (1.5/q*D_eps_p_eq);
      for ( i=1; i<=3       ; i++ ) e_elas1(i) -= alpha*S(i);
      alpha *= 2.0;
      for ( i=4; i<=xstrsize; i++ ) e_elas1(i) -= alpha*S(i);
   }

   if ( nu == 0.5 )
   { cerr << "nu = 0.5, going to divide by 0; error" << endl;
     throw( EXIT_FAILURE ); }

   K = E/(3*(1-2*nu));
   p1 = p0 - K*Deps_V;
}

void VonMises_Point::finish_increment()
{
   e_elas0 = e_elas1;
   eq0 = eq1;
   p0 = p1;

   int xstrsize = ExpandedStressSize( stress_type );
   double e_base = square(e_elas0(1)) + square(e_elas0(2)) + square(e_elas0(3));
   for ( int i=4; i<=xstrsize; i++ ) e_base += 0.5*square(e_elas0(i));

   e_tilde = sqrt( 1.5*e_base );
   strain_bar.clear();
   prev_dstrain.clear();
}

double VonMises_Point::get_eps_plast_eq( bool &has_epq ) const
{
   has_epq = true;
   return eq1;
}

void VonMises_Point::GetRestartData( CollectArray &data ) const
{
   data.reserve( e_elas0.size()+2 );
   data.add(e_elas0);
   data.add( eq0 );
   data.add( p0 );

   if ( plastic ) data.add( 1.0 );
   else data.add( 0.0 );
}

void VonMises_Point::PutRestartData( const CollectArray &data )
{
   data.extract(e_elas0);
   eq0 = data.extract();
   p0 = data.extract();

   e_elas1 = e_elas0;
   p1 = p0;
   eq1 = eq0;
   double dummy( data.extract() );

   plastic = ( dummy > 0.99 && dummy < 1.01 );

   if ( plastic )
   {
      int xstrsize = ExpandedStressSize( stress_type );
      double e_base = square(e_elas0(1)) + square(e_elas0(2)) + square(e_elas0(3));
      for ( int i=4; i<=xstrsize; i++ ) e_base += 0.5*square(e_elas0(i));

      e_tilde = sqrt( 1.5*e_base );
   }
}

ThermoPlastic_Point::ThermoPlastic_Point( ThermoPlastic_type *mat, StressType st )
                               :VonMises_Point( mat, st ),LinearThermal_Point( mat, st )
{
   thermal_strain_old = Delta_eps_th = 0.0;
}

ThermoPlastic_Point::~ThermoPlastic_Point(){}

void ThermoPlastic_Point::set_strain_increment( const Vector& strain_increment )
{
   Vector elpl_strain_incr(strain_increment);

   elpl_strain_incr(1) -= Delta_eps_th;
   switch ( stress_type )
   {
      case axi:
      case gen: elpl_strain_incr(3) -= Delta_eps_th;
      case pss:
      case psn: elpl_strain_incr(2) -= Delta_eps_th;
      case uni: break;
   }

   VonMises_Point::set_strain_increment( elpl_strain_incr );
}

void ThermoPlastic_Point::set_potential_increment( double Tinc )
{
   Delta_eps_th = Tinc * get_expansion_coefficient();
   LinearThermal_Point::set_potential_increment( Tinc );
}

void ThermoPlastic_Point::finish_increment()
{
   thermal_strain_old += Delta_eps_th;
   Delta_eps_th = 0.0;
   VonMises_Point::finish_increment();
   LinearThermal_Point::finish_increment();
}

void ThermoPlastic_Point::ExpandEps( Vector& dstrain, Vector& ddstrain, Vector& strainbar )
{
   FeaTure::Structural_Point::ExpandEps( dstrain, ddstrain, strainbar );

   if ( stress_type == psn ) dstrain(3) -= Delta_eps_th;
}

double ThermoPlastic_Point::get_expansion_coefficient() const
{
#ifndef NOVARIABLERETURN
   return dynamic_cast<ThermoPlastic_type *>
                     (LinearThermal_Point::parameters)->get_alpha();
#else
   return ((ThermoPlastic_type *)LinearThermal_Point::parameters)->get_alpha();
#endif
}

void ThermoPlastic_Point::GetRestartData( CollectArray &data ) const
{
   VonMises_Point::GetRestartData( data );
   LinearThermal_Point::GetRestartData( data );
   data.add( thermal_strain_old );
}

void ThermoPlastic_Point::PutRestartData( const CollectArray &data )
{
   VonMises_Point::PutRestartData( data );
   LinearThermal_Point::PutRestartData( data );
   thermal_strain_old = data.extract();
}

VonMises_type::VonMises_type( int matnr ):FeaTure::Structural_type()
{
   nr = matnr;
   Young = 0;
   Poisson = 0;
   eps0 = 0;
   C = 0;
   n = 0;
}

VonMises_type::~VonMises_type(){}

#ifdef NOVARIABLERETURN
FeaTure::Material_Point* VonMises_type::new_Material_Point( StressType st )
#else
VonMises_Point* VonMises_type::new_Material_Point( StressType st )
#endif
{
   return new VonMises_Point(this,st);
}

int VonMises_type::set_Young( double young )
{
   Young = young;

   if ( Young <= 0.0 )
   {
      cerr << "Youngs modulus <= 0; this is not allowed" << endl;
      return 1;
   }
   return 0;
}

double VonMises_type::get_Young() const
{
   return Young;
}

int VonMises_type::set_Poisson( double poisson )
{
   Poisson = poisson;
   if ( Poisson < 0 || Poisson > 0.5 )
   {
      cerr << "Poisson's ratio must be between 0 and 0.5; input error" << endl;
      return 1;
   }
   return 0;
}

double VonMises_type::get_Poisson() const
{
   return Poisson;
}

int VonMises_type::set_yield( double s0 )
{
   if ( s0 <= 0 )
   {
      cerr << "The initial yield stress must be greater than 0; input error" << endl;
      return 1;
   }
   assert( C!=0 && n!=0 );
   eps0 = pow(s0/C,1/n);
   return 0;
}

double VonMises_type::get_eps0() const
{
   return eps0;
}

int VonMises_type::set_nadai_C( double c )
{
   C = c;
   if ( C <= 0 )
   {
      cerr << "The multiplier C must be greater than 0; input error" << endl;
      return 1;
   }
   return 0;
}

double VonMises_type::get_nadai_C() const
{
   return C;
}

int VonMises_type::set_nadai_n( double nadai_n )
{
   n = nadai_n;
   if ( n<0 || n>1 )
   {
      cerr << "The exponent n must be between 0 and 1; input error" << endl;
      return 1;
   }
   return 0;
}

double VonMises_type::get_nadai_n() const
{
   return n;
}

bool VonMises_type::NeedExpandStrain( StressType st ) const
{
   if ( st == pss || st == psn ) return true;
   else return false;
}

ThermoPlastic_type::ThermoPlastic_type( int matnr )
                   :VonMises_type( matnr ),LinearThermal_type(matnr)
{}

ThermoPlastic_type::~ThermoPlastic_type(){}

#ifdef NOVARIABLERETURN
FeaTure::Material_Point* ThermoPlastic_type::new_Material_Point( StressType st )
#else
ThermoPlastic_Point* ThermoPlastic_type::new_Material_Point( StressType st )
#endif
{
   return new ThermoPlastic_Point(this,st);
}

int ThermoPlastic_type::set_alpha( double al )
{
   alpha = al;

   if ( alpha <= 0.0 )
   {
      cerr << "expansion coefficient <= 0; this is not allowed" << endl;
      return 1;
   }
   return 0;
}
