// Copyright (c) 1999, A.H. van den Boogaard

#ifndef CONTROL_H
#define CONTROL_H

#include <string>
#include <deque>

#include "sysmat.h"
#include "ftmatrix.h"
#include "stpselect.h"
#include "postproc.h"
#include "ftexcept.h"

namespace FeaTure
{

template <class DomainT>
class Control
{
   protected:
      bool        has_run;
      PostProc    postproc;
      std::string file_basename;

      Control(){ has_run = false; }

   public:
      virtual ~Control(){}
      virtual void solve( DomainT& model, std::ostream &out, std::ostream &post );

      virtual void initialize( DomainT& model, std::ostream &out )=0;
      virtual bool continuate( DomainT& model, std::ostream &out,
                               std::ostream &post, int steps )=0;

      void set_basename( const std::string& filename )
         { file_basename = filename; }
      void set_postproc( const PostProc& post ){ postproc = post; }
      bool need_post_file(){ return postproc.any_selected(); }
};

template <class DomainT>
class IncrementalControl:public Control<DomainT>
{
   protected:
      enum StartUp { start, restart, contin };

      struct LoadBlock
      {
         int load_set;
         double size;
         int nr_increments;
      };

      IncrementalControl();

      StartUp     startup;
      int         max_incr, incr_nr;
      double      time_increment;
      double      time0, time1;
      Vector      rhs;
      double      Rayleigh_alpha, Rayleigh_beta;
      bool        use_damping;
      StepSelect  save_restart_file;
      int         max_resize;
      std::string restartfilename;
      std::deque<LoadBlock> load_blocks;

      void init_incremental( DomainT& model );
      virtual void solve_increment( DomainT& model, std::ostream& out,
                                    int loadset, double size )=0;
      virtual void log_increment_nr( std::ostream &out );

   public:
      ~IncrementalControl(){}

      enum WhatFraction{ STEPS };
      void set_increments( int set, double size, int steps );
      void set_time_increment( double dtime );
      float get_fraction_executed( int flag );
      void set_Rayleigh_damping( double alpha, double beta );
      void set_max_resize( int i ){ max_resize = i; }
      bool is_last(){ return max_incr == incr_nr; }
      void set_restart( const std::string& filename )
         { startup = restart; restartfilename = filename; }

      virtual bool continuate( DomainT& model, std::ostream &out,
                               std::ostream &post, int steps );
      virtual void WriteRestartData( DomainT& model );
      virtual void ReadRestartData( DomainT& model );

      void set_restart_select( const StepSelect& restart_select )
           { save_restart_file = restart_select; }
};

template <class DomainT>
class IncrementalIterativeControl:public IncrementalControl<DomainT>
{
   protected:
      int         iteration_method, max_iter, nr_iter;
      double      EPS_UNB;
      double      load_norm, unbalance_norm;

      IncrementalIterativeControl();

      virtual void calculate_norms();
      virtual bool converged() const;
      virtual void solve_iteration( SystemMat &Systemptr, DomainT& model );

   public:
      ~IncrementalIterativeControl();

      void set_max_iter( int maxiter );
      void set_unbalance_crit( double eps );
      void set_iteration_method( int method );
};

class not_converged_error:public feature_error
{
   public:
      not_converged_error( const std::string& what_arg ): feature_error(what_arg){}
};

class divergence_error:public not_converged_error
{
   public:
      divergence_error( const std::string& what_arg ): not_converged_error(what_arg){}
};

} // namespace FeaTure

#endif /* CONTROL_H */
