// Copyright (c) 1999, A.H. van den Boogaard

#include "domain.h"
#include "ftexcept.h"
#include <algorithm>

template <class Element, class Node>
FeaTure::Domain<Element,Node>::Domain()
{
   nr_free=0; nr_fixed=0; nr_prescribed=0;
   symmetric = true;
}

template <class Element, class Node>
FeaTure::Domain<Element,Node>::~Domain()
{
   for ( ElmIterator elm=elements.begin(); elm!=elements.end(); elm++ )
      delete (*elm);

   for ( NodeIterator node=nodes.begin(); node!=nodes.end(); node++ )
      delete (*node);

   for ( MaterialIterator mat=materials.begin(); mat!=materials.end(); mat++ )
      delete (*mat);

   for ( GeomIterator geom=geometries.begin(); geom!=geometries.end(); geom++ )
      delete (*geom);

   for ( BoundCondIterator bc=loads.begin(); bc!=loads.end(); bc++ )
      delete (*bc);
}

template <class Element, class Node>
void FeaTure::Domain<Element,Node>::add_element( const Element_ptr new_element )
{
   elements.insert( elements.end(), new_element );
   symmetric = symmetric && new_element->symmetric();
}

template <class Element, class Node>
void FeaTure::Domain<Element,Node>::add_node( const Node_ptr new_node )
{
   nodes.insert( nodes.end(), new_node );
}

template <class Element, class Node>
void FeaTure::Domain<Element,Node>::add_load_set( LoadSet *load )
{
   loads.insert( loads.end(), load );
}

template <class Element, class Node>
void FeaTure::Domain<Element,Node>::add_material( Material_type *mat )
{
   materials.insert( materials.end(), mat );
}

template <class Element, class Node>
void FeaTure::Domain<Element,Node>::add_geometry( Geometry *geom )
{
   geometries.insert( geometries.end(), geom );
}

template <class Type>
class IsKey
{
   public:
      IsKey( int Key ):testValue(Key){}

      bool operator() ( Type* type )
         { return ( type->get_nr() == testValue ); }

   private:
      int testValue;
};

template <class Element, class Node>
FeaTure::Material_type*
         FeaTure::Domain<Element,Node>::get_mat_ptr( int nr ) const
{
   ConstMaterialIterator where(
     std::find_if( materials.begin(), materials.end(),
                   IsKey<Material_type>( nr ) ) );
   if ( where != materials.end() )
     return *where;
   else
     throw( not_found( "unknown material number" ) );
}

template <class Element, class Node>
FeaTure::Geometry* FeaTure::Domain<Element,Node>::get_geom_ptr( int nr ) const
{
   ConstGeomIterator where =
     std::find_if( geometries.begin(), geometries.end(),
                   IsKey<Geometry>( nr ) );
   if ( where != geometries.end() )
     return *where;
   else
     throw( not_found( "unknown geometry number" ) );
}

template <class Element, class Node>
Node* FeaTure::Domain<Element,Node>::get_node_ptr( int nr ) const
{
   ConstNodeIterator where =
     std::find_if( nodes.begin(), nodes.end(), IsKey<Node>( nr ) );
   if ( where != nodes.end() )
     return *where;
   else
     throw( not_found( "unknown node number" ) );
}

template <class Element, class Node>
Element* FeaTure::Domain<Element,Node>::get_element_ptr( int nr ) const
{
   ConstElmIterator where =
     std::find_if( elements.begin(), elements.end(), IsKey<ElementBase>( nr ) );
   if ( where != elements.end() )
     return *where;
   else
     throw( not_found( "unknown element number" ) );
}

template <class Element, class Node>
void FeaTure::Domain<Element,Node>::create_element_dofs()
{
   for ( ElmIterator elm=elements.begin(); elm!=elements.end(); elm++ )
      (*elm)->add_dofs_to_nodes();
}

template <class Element, class Node>
void FeaTure::Domain<Element,Node>::number_dofs()
{
   for ( NodeIterator node=nodes.begin(); node!=nodes.end(); node++ )
      (*node)->make_global_dofs( nr_free, nr_fixed, nr_prescribed );
}

template <class Element, class Node>
void FeaTure::Domain<Element,Node>::make_global_dofs()
{
   create_element_dofs();
   number_dofs();
   make_transmats();
}

template <class Element, class Node>
void FeaTure::Domain<Element,Node>::make_transmats()
{
   /* Set up element dofs */
   for ( ElmIterator elm=elements.begin(); elm!=elements.end(); elm++ )
      (*elm)->CreateGlobalDofs();
      
   return;
}

template <class Element, class Node>
void FeaTure::Domain<Element,Node>::init_system( SystemMat &System )
{
   /* Set up element dofs and initialize sparse system matrix */
   for ( ElmIterator elm=elements.begin(); elm!=elements.end(); elm++ )
   {
      (*elm)->InitStiff_in_Global( &System );
   }
   return;
}

template <class Element, class Node>
void FeaTure::Domain<Element,Node>::make_system( SystemMat &System,
                                                 Vector &rhs )
{
   /* Assemble system matrix from elements
      and take care of prescribed dofs     */

   System.clear();
   rhs.clear();
   for ( ElmIterator elm=elements.begin(); elm!=elements.end(); elm++ )
      (*elm)->PutStiff_and_RHS_in_Global( &System, rhs );

   return;
}

template <class Element, class Node>
void FeaTure::Domain<Element,Node>::make_mass_system(
                                     SystemMat &System, bool lumped )
{
   /* Assemble system mass matrix from elements */

   System.clear();
   for ( ElmIterator elm=elements.begin(); elm!=elements.end(); elm++ )
      (*elm)->PutMass_in_Global( &System, lumped );

   return;
}

template <class Element, class Node>
void FeaTure::Domain<Element,Node>::update_system( SystemMat &System,
                                                   Vector &rhs )
{
   /* Assemble system matrix from elements */

   System.clear();
   for ( ElmIterator elm=elements.begin(); elm!=elements.end(); elm++ )
      (*elm)->PutStiff_in_Global( &System, rhs );

   return;
}

template <class Element, class Node>
void FeaTure::Domain<Element,Node>::get_load_vector(
                                    SystemVector& load_vector ) const
{
   load_vector.resize( nr_free, nr_prescribed, nr_fixed );
   load_vector.clear();

   SystemVector local_load( nr_free, nr_prescribed, nr_fixed );

   for ( ConstBoundCondIterator bc=loads.begin(); bc!=loads.end(); bc++ )
   {
      (*bc)->get_load_vector( local_load );
      load_vector += local_load;
   }
}

template <class Element, class Node>
void FeaTure::Domain<Element,Node>::set_load_size( int set, double size )
{
   ConstBoundCondIterator where =
     std::find_if( loads.begin(), loads.end(), IsKey<LoadSet>( set ) );
   if ( where != loads.end() )
     (*where)->set_load_size( size );
   else
     throw( not_found( "unknown load set number" ) );
}

template <class Element, class Node>
void FeaTure::Domain<Element,Node>::add_load_size( int set, double size )
{
/*   for ( BoundCondIterator bc=loads.begin(); bc!=loads.end(); bc++ )
      cout << "load set " << (*bc)->get_nr() << endl; */

   ConstBoundCondIterator where =
     std::find_if( loads.begin(), loads.end(), IsKey<LoadSet>( set ) );
   if ( where != loads.end() )
     (*where)->add_load_size( size );
   else
     throw( not_found( "unknown load set number" ) );
}

template <class Element, class Node>
void FeaTure::Domain<Element,Node>::make_LHS( SystemVector &lhs ) const
{
   lhs.resize( nr_free, nr_prescribed, nr_fixed );
   lhs.clear();

   for ( ConstElmIterator elm=elements.begin(); elm!=elements.end(); elm++ )
      (*elm)->PutLHS_in_Global( lhs );
}

template <class Element, class Node>
void FeaTure::IncDomain<Element,Node>::finish_increment()
{
   // prepare model for the next increment
   // if start and end data is stored, copy end to start
   // if start and increments are stored, add incr. to total and reset incr.

   add_increments_to_totals();

   for ( ElmIterator elm=this->elements.begin();
                     elm!=this->elements.end(); elm++ )
      (*elm)->finish_increment();
}

template <class Element, class Node>
void FeaTure::IncDomain<Element,Node>::iteration_update()
{
   for ( ElmIterator elm=this->elements.begin();
                     elm!=this->elements.end(); elm++ )
      (*elm)->iteration_update();

   return;
}

template <class Element, class Node>
void FeaTure::IncDomain<Element,Node>::reset_increments()
{
   for ( NodeIterator node=this->nodes.begin();
                      node!=this->nodes.end(); node++ )
      (*node)->reset_increments();
}

template <class Element, class Node>
void FeaTure::IncDomain<Element,Node>::add_increments_to_totals()
{
   for ( NodeIterator node=this->nodes.begin();
                      node!=this->nodes.end(); node++ )
      (*node)->add_increments_to_totals();
}

template <class Element, class Node>
void FeaTure::IncDomain<Element,Node>::add_to_increments(
               const Vector &free, const Vector &prescribed )
{
   for ( NodeIterator node=this->nodes.begin();
                      node!=this->nodes.end(); node++ )
      (*node)->add_to_increments( free, prescribed );
}

template <class Element, class Node>
void FeaTure::IncDomain<Element,Node>::add_to_increments( const Vector &free )
{
   for ( NodeIterator node=this->nodes.begin();
                      node!=this->nodes.end(); node++ )
      (*node)->add_to_increments( free );
}

template <class Element, class Node>
void FeaTure::IncDomain<Element,Node>::get_incremental_solution_vector(
                                                      Vector &increments_free )
{
   increments_free.resize( this->get_nr_free() );

   for ( NodeIterator node=this->nodes.begin();
                      node!=this->nodes.end(); node++ )
      (*node)->get_incremental_solution_vector( increments_free );
}

template <class Element, class Node>
void FeaTure::IncDomain<Element,Node>::WriteRestartData(
                                                  std::ostream &restart ) const
{
   for ( ConstNodeIterator node=this->nodes.begin();
                           node!=this->nodes.end(); node++ )
      (*node)->WriteRestartData( restart );

   for ( ConstElmIterator elm=this->elements.begin();
                          elm!=this->elements.end(); elm++ )
      (*elm)->WriteRestartData( restart );

   for ( ConstBoundCondIterator bc=this->loads.begin();
                                bc!=this->loads.end(); bc++ )
      (*bc)->WriteRestartData( restart );
}

template <class Element, class Node>
void FeaTure::IncDomain<Element,Node>::ReadRestartData( std::istream &restart )
{
   for ( ConstNodeIterator node=this->nodes.begin();
                           node!=this->nodes.end(); node++ )
      (*node)->ReadRestartData( restart );

   for ( ConstElmIterator elm=this->elements.begin();
                          elm!=this->elements.end(); elm++ )
      (*elm)->ReadRestartData( restart );

   for ( ConstBoundCondIterator bc=this->loads.begin();
                                bc!=this->loads.end(); bc++ )
      (*bc)->ReadRestartData( restart );
}


