// Copyright (c) 1999, A.H. van den Boogaard

#ifndef DOMAIN_H
#define DOMAIN_H

#include <list>
#include <string>

#include "loads.h"
#include "material.h"
#include "geometry.h"

namespace FeaTure
{

template <class Element, class Node>
class Domain
{
   protected:
      typedef Element* Element_ptr;
      typedef Node*    Node_ptr;

      typedef std::list<Element_ptr> ElementList;
      typedef typename std::list<Element_ptr>::iterator ElmIterator;
      typedef typename std::list<Element_ptr>::const_iterator ConstElmIterator;

      typedef std::list<Node_ptr> NodeList;
      typedef typename std::list<Node_ptr>::iterator NodeIterator;
      typedef typename std::list<Node_ptr>::const_iterator ConstNodeIterator;

      typedef std::list<Material_type*> Material_types;
      typedef typename std::list<Material_type*>::iterator MaterialIterator;
      typedef typename std::list<Material_type*>::const_iterator ConstMaterialIterator;

      typedef std::list<Geometry*> Geometries;
      typedef typename std::list<Geometry*>::iterator GeomIterator;
      typedef typename std::list<Geometry*>::const_iterator ConstGeomIterator;

      typedef std::list<LoadSet*> BoundConds;
      typedef typename std::list<LoadSet*>::iterator BoundCondIterator;
      typedef typename std::list<LoadSet*>::const_iterator ConstBoundCondIterator;

      NodeList       nodes;
      ElementList    elements;
      std::string    title;
      BoundConds     loads;

   private:
      Material_types materials;
      Geometries     geometries;

      int            nr_free, nr_fixed, nr_prescribed;
      bool           symmetric;

   protected:
      void create_element_dofs();
      virtual void number_dofs();
      void make_transmats();

   public:
      Domain();
      virtual ~Domain();
      void add_element( const Element_ptr );
      void add_node( const Node_ptr );
      void add_load_set( LoadSet *load );
      void set_title( const std::string &text ){ title = text; }
      void add_material( Material_type *mat );
      void add_geometry( Geometry *geom );
      void make_global_dofs();

      Material_type* get_mat_ptr( int matnr ) const;
      Geometry* get_geom_ptr( int geomnr ) const;
      Node_ptr get_node_ptr( int nodenr ) const;
      Element_ptr get_element_ptr( int elmnr ) const;
      int  get_nr_nodes() const { return nodes.size(); }
      int  get_nr_elements() const { return elements.size(); }
      int  get_nr_free() const { return nr_free; }
      int  get_nr_prescribed() const { return nr_prescribed; }
      int  get_nr_fixed() const { return nr_fixed; }
      bool need_non_symmetric() const { return !symmetric; }

      void init_system( SystemMat &system );
      void make_LHS( SystemVector &f_int ) const;
      void make_system( SystemMat &System, Vector &rhs );
      void make_mass_system( SystemMat &System, bool lumped );
      void update_system( SystemMat &System, Vector &rhs );

      void get_load_vector(SystemVector& load_vector ) const;
      void set_load_size( int set, double size );
      void add_load_size( int set, double size );

      virtual void print_results( std::ostream& /*out*/ ) const {}
};

template <class Element, class Node>
class IncDomain:public Domain<Element,Node>
{
   protected:
      typedef Element* Element_ptr;
      typedef Node*    Node_ptr;

      typedef std::list<Element_ptr> ElementList;
      typedef typename std::list<Element_ptr>::iterator ElmIterator;
      typedef typename std::list<Element_ptr>::const_iterator ConstElmIterator;

      typedef std::list<Node_ptr> NodeList;
      typedef typename std::list<Node_ptr>::iterator NodeIterator;
      typedef typename std::list<Node_ptr>::const_iterator ConstNodeIterator;

      typedef std::list<Material_type*> Material_types;
      typedef typename std::list<Material_type*>::iterator MaterialIterator;
      typedef typename std::list<Material_type*>::const_iterator ConstMaterialIterator;

      typedef std::list<Geometry*> Geometries;
      typedef typename std::list<Geometry*>::iterator GeomIterator;
      typedef typename std::list<Geometry*>::const_iterator ConstGeomIterator;

      typedef std::list<LoadSet*> BoundConds;
      typedef typename std::list<LoadSet*>::iterator BoundCondIterator;
      typedef typename std::list<LoadSet*>::const_iterator ConstBoundCondIterator;

   public:
      void finish_increment();
      void iteration_update();
      void add_increments_to_totals();
      void reset_increments();
      void add_to_increments( const Vector &free, const Vector &prescribed );
      void add_to_increments( const Vector &free );
      void get_incremental_solution_vector( Vector &increments_free );

      void WriteRestartData( std::ostream &restart ) const;
      void ReadRestartData( std::istream &restart );
 };

} // namespace FeaTure

#endif /* DOMAIN_H */
