// Copyright (c) 1999, A.H. van den Boogaard

#ifndef LOADS_H
#define LOADS_H

#include "nodes.h"
#include "elembase.h"
#include "ftmatrix.h"
#include "sysvec.h"
#include <list>
#include <vector>

namespace FeaTure
{

class Nodal_Loads
{
   private:
      typedef Node* NodePtr;

      class Load_dir
      {
         public:
            Load_dir(){}
            Dof::Load type;
            Direction dir;
            double    magnitude;
      };

      std::vector<Load_dir> loads;
      std::vector<NodePtr>  nodes;

   public:
      Nodal_Loads();
      Nodal_Loads( int nrloads, Load_dir *lods,
                   int nrnodes, NodePtr *nods );
      ~Nodal_Loads();

      void add_load( Dof::Load type, Direction dir, double mag );
      void add_load( Dof::Load type, double mag );
      void add_node( NodePtr node  );
      int get_nr_loads() const;
      Dof::Load get_type( int load_nr ) const;
      Direction get_dir(  int load_nr ) const;
      double get_magnitude( int load_nr ) const;
      int get_nr_nodes() const;
      NodePtr get_node( int node_nr ) const;

      void print( std::ostream& out ) const;
      void add_to_rhs( SystemVector& rhs ) const;
};

class Nodal_Prescribed
{
   private:
      typedef Node* NodePtr;

      class Prescr_dir
      {
         public:
            Prescr_dir(){}
            Dof::Type type;
            Direction dir;
            double    magnitude;
      };

      std::vector<Prescr_dir> loads;
      std::vector<NodePtr>    nodes;

   public:
      Nodal_Prescribed();
      Nodal_Prescribed( int nrloads, Prescr_dir *lods,
                        int nrnodes, NodePtr *nods );
      ~Nodal_Prescribed();

      void add_load( Dof::Type type, Direction dir, double mag );
      void add_load( Dof::Type type, double mag );
      void add_node( NodePtr node );
      int get_nr_loads() const;
      Dof::Type get_type( int load_nr ) const;
      Direction get_dir(  int load_nr ) const;
      double get_magnitude( int load_nr ) const;
      int get_nr_nodes() const;
      NodePtr get_node( int node_nr ) const;

      void print( std::ostream& out ) const;
      void add_to_prescribed( double inc_size ) const;
};

class Pressure_Loads
{
   private:
      typedef ElementBase* Element_ptr;

      double factor;

      class PresLod
      {
         public:
            ElementBase* elm_ptr;
            int    surface_nr;
            double magnitude;

            PresLod():elm_ptr(NULL),surface_nr(0),magnitude(0.0){}
            PresLod(ElementBase* e, int s, double m):elm_ptr(e),surface_nr(s),
                                 magnitude(m){}
            ~PresLod(){}
            int number() const { return surface_nr; }
            /* std::list<PresLod> seems to need the definition of == and < */
            bool operator==(const PresLod& P)const{ return elm_ptr==P.elm_ptr &&
                          surface_nr==P.surface_nr; }
            bool operator<(const PresLod &P)const
                          { return magnitude < P.magnitude;}
      };

      std::list<PresLod> pres_list;
      typedef std::list<PresLod>::const_iterator ConstPresIterator;
      typedef std::list<PresLod>::iterator PresIterator;

   public:
      Pressure_Loads(){ factor=0; }
      ~Pressure_Loads(){}

      void add_load( Element_ptr elmptr, int surface, double pres )
      {
         PresLod item(elmptr,surface,pres);
         pres_list.push_back(item);
      }

      void set_factor( double fac ){ factor = fac; }
      void add_factor( double fac ){ factor += fac; }

      void add_to_vec( SystemVector& load_vec ) const;
};

class LoadSet
{
   private:
      int    loadset_nr;
      double multiplier;

      std::vector<Nodal_Loads*>      nodal_loads;
      std::vector<Nodal_Prescribed*> nodal_prescribed;
      std::vector<Pressure_Loads*>   pressure_loads;

      void add_to_prescribed( double inc_size ) const;

   public:
      LoadSet();
      LoadSet( int nr );
      ~LoadSet();

      int get_nr() const { return loadset_nr; }
      void set_nr( int nr ){ loadset_nr = nr; }

      void add_load( Nodal_Loads *load );
      void add_load( Nodal_Prescribed *load );
      void add_load( Pressure_Loads *load );
      void print( std::ostream& out ) const;

      void get_load_vector( SystemVector& load_vec ) const;

      void set_load_size( double size );
      void add_load_size( double size );
      void WriteRestartData( std::ostream& restart ) const;
      void ReadRestartData( std::istream& restart );
};

} // namespace FeaTure

#endif // LOADS_H
