// Copyright (c) 1999, A.H. van den Boogaard

#include "nodes.h"
#include "ftexcept.h"

using FeaTure::Vector;
using FeaTure::CartVector;
using FeaTure::SystemVector;
using FeaTure::Direction;
using FeaTure::Location;
using FeaTure::Node;
using FeaTure::IncNode;
using FeaTure::Dof;
using FeaTure::dof_sets;
using FeaTure::Dof_ptr;

/* static data member */
double Node::eps_direct = 1.0e-6;

Node::Node()
{
   nr_nodal_dofs = 0;
   dofs = NULL;
   node_nr = 0;
}

Node::Node( int nr )
{
   nr_nodal_dofs = 0;
   dofs = NULL;
   node_nr = nr;
}

Node::Node( int nr, double x, double y, double z ):coordinates(x,y,z)
{
   nr_nodal_dofs = 0;
   dofs = NULL;
   node_nr = nr;
}

Node::~Node()
{
   int idof;

   for ( idof = 0; idof<nr_nodal_dofs; idof++ )
      delete( dofs[idof] );

   if ( dofs != NULL ) std::free( dofs );
}

void Node::set_nr( int nr )
{
   node_nr = nr;
}

int Node::get_nr() const
{
   return node_nr;
}

void Node::add_dof( Dof_ptr plus_dof )
{
   /* Only substitute malloc/realloc by vector if you are sure that vector
      does not use too much extra space */
   nr_nodal_dofs++;
   if ( dofs == NULL ) dofs = (Dof_ptr*) std::malloc( sizeof(Dof_ptr) );
   else dofs = (Dof_ptr*) std::realloc( dofs, nr_nodal_dofs*sizeof(Dof_ptr) );

   dofs[nr_nodal_dofs-1] = plus_dof;
}

Dof_ptr Node::get_dof( int idof ) const
{
   if ( idof >= nr_nodal_dofs ) throw dimension_error( "Node::get_dof( int )" );
   return dofs[idof];
}

void Node::get_dof( Dof::Type phys_type, Direction dir, dof_sets *SystDofs ) const
{
/* get a combination of degrees of freedom that represent the type and
   direction of the arguments, put it in SystDofs */

   int        idof;
   CartVector dofdir;
   double     inp;

   for ( idof=0; idof<nr_nodal_dofs; idof++ )
   {
      if ( phys_type == dofs[idof]->get_dof_type() )
      {
         if ( dofs[idof]->is_vector() )
         {
            dofdir = dofs[idof]->get_direction();
            inp = dir*dofdir;
            dir -= inp*dofdir;
         }
         else inp = 1.0;

         if ( std::fabs(inp) > eps_direct )
            SystDofs->add_set( dofs[idof], inp );

         if ( dofs[idof]->is_scalar() || dir.normsq() < eps_direct ) break;
      }
   }
}

void Node::set_location( Location loc )
{
   coordinates = loc;
}

void Node::set_location( double x, double y, double z )
{
   coordinates.set_location( x, y, z );
}

Location Node::get_location() const
{
   return coordinates;
}

void Node::get_location( double& x, double& y, double& z ) const
{
   coordinates.get_location( x, y, z );
}

void Node::get_current_location( double& x, double& y, double& z ) const
{
   double     value;
   Location   P, U(0,0,0), dir;

   for ( int idof=0; idof<nr_nodal_dofs; idof++ )
   {
      if ( dofs[idof]->get_dof_type() == Dof::u )
      {
         value = dofs[idof]->get_total();

         dir = dofs[idof]->get_direction();
         U += dir*value;
      }
   }
   P = coordinates+U;

   P.get_location( x, y, z );
}

void Node::draw_location( double& x, double& y, double& z, double magnif ) const
{
   double     value;
   Location   P, U(0,0,0), dir;

   for ( int idof=0; idof<nr_nodal_dofs; idof++ )
   {
      if ( dofs[idof]->get_dof_type() == Dof::u )
      {
         value = dofs[idof]->get_total();

         dir = dofs[idof]->get_direction();
         U += dir*value*magnif;
      }
   }
   P = coordinates+U;

   P.get_location( x, y, z );
}

int Node::create_dof( Dof::dof_fixed fix_type, Dof::Type phys_type,
                      const Direction& dir )
{
   int        idof;
   double     inp;
   CartVector newdir, dofdir;

   if ( is_vector_dof( phys_type ) ) newdir = dir;

   for ( idof=0; idof<nr_nodal_dofs; idof++ )
   {
      if ( phys_type == dofs[idof]->get_dof_type() )
      {
         if ( dofs[idof]->is_vector() )
         {
            dofdir = dofs[idof]->get_direction();
            inp = newdir*dofdir;
            newdir -= inp*dofdir;
         }
         else return(1);
      }
   }

   if ( is_scalar_dof( phys_type ) || newdir.normsq() > eps_direct )
   {    // create new degree of freedom
      itincDof *newdof = new( itincDof );
      newdof->set_dof_fixed( fix_type );
      newdof->set_dof_type( phys_type );
      if ( is_vector_dof( phys_type ) ) newdof->set_direction( newdir );

      add_dof( newdof );
      return(0);
   }
   else return(1);
}

void Node::make_global_dofs( int& nr_free, int& nr_fixed, int& nr_prescribed )
{
   int idof;
   for ( idof=0; idof<nr_nodal_dofs; idof++ )
   {
      if ( dofs[idof]->is_free() )
         dofs[idof]->set_global_dof( ++nr_free );
      else if ( dofs[idof]->is_prescribed() )
         dofs[idof]->set_global_dof( ++nr_prescribed );
      else if ( dofs[idof]->is_fixed() )
         dofs[idof]->set_global_dof( ++nr_fixed );
   }
}

void Node::add_to_rhs( Dof::Load type, Direction dir, double magnitude,
                       SystemVector& rhs ) const
{
   int        idof;
   Dof::Type  phys_type;
   double     inp;
   CartVector dofdir;

   phys_type = conjugate_type( type );

   for ( idof=0; idof<nr_nodal_dofs; idof++ )
   {
      if ( phys_type == dofs[idof]->get_dof_type() )
      {
         if ( dofs[idof]->is_free() )
         {
            if ( dofs[idof]->is_vector() )
            {
               dofdir = dofs[idof]->get_direction();
               inp = dir*dofdir;
               dir -= inp*dofdir;
               rhs(*(dofs[idof])) += magnitude*inp;
            }
            else
            {
               rhs(*(dofs[idof])) += magnitude;
               return;
            }
         }
      }
   }
   if ( dir.normsq() > eps_direct )
   {
     std::cerr << "load lost; type or direction not matched by independent dofs"
               << std::endl;
     return;
   }
   return;
}

void Node::add_to_prescribed( Dof::Type type, Direction dir, double magnitude ) const
{
   int        idof;
   double     inp;
   CartVector dofdir;

   for ( idof=0; idof<nr_nodal_dofs; idof++ )
   {
      if ( type == dofs[idof]->get_dof_type() )
      {
         if ( dofs[idof]->is_prescribed() )
         {
            if ( dofs[idof]->is_vector() )
            {
               dofdir = dofs[idof]->get_direction();
               inp = dir*dofdir;
               dir -= inp*dofdir;
               dofs[idof]->add_to_increment( magnitude*inp );
            }
            else
            {
               dofs[idof]->add_to_increment( magnitude );
               return;
            }
         }
      }
   }
   if ( dir.normsq() > eps_direct )
   {
      std::cerr << "prescribed displacement lost;"
                << " type or direction not matched by prescribed dofs"
                << std::endl;
      return;
   }
   return;
}

CartVector Node::get_total_displacement() const
{
   double     value;
   CartVector U(0,0,0);
   Direction  dir;

   for ( int idof=0; idof<nr_nodal_dofs; idof++ )
   {
      if ( dofs[idof]->get_dof_type() == Dof::u )
      {
         value = dofs[idof]->get_total();

         dir = dofs[idof]->get_direction();
         U += dir*value;
      }
   }

   return U;
}

double Node::get_temperature() const
{
   for ( int idof=0; idof<nr_nodal_dofs; idof++ )
   {
      if ( dofs[idof]->get_dof_type() == Dof::t )
         return dofs[idof]->get_total();
   }

   return 0.0;
}

void Node::WriteRestartData( std::ostream &restart ) const
{
   restart << 'N' << get_nr() << " ";
   for ( int idof=0; idof<nr_nodal_dofs; idof++ )
      restart << dofs[idof]->get_total() << " ";

   restart << std::endl;
}

void Node::ReadRestartData( std::istream &restart)
{
   double value;

   char c;
   int  nr;

   restart >> c >> nr;
   if ( c != 'N' || nr != get_nr() )
      throw file_error( "corrupted restart file" );

   for ( int idof=0; idof<nr_nodal_dofs; idof++ )
   {
      restart >> value;
      dofs[idof]->set_total( value );
   }
}

void IncNode::add_increments_to_totals()
{
   for ( int idof=0; idof<nr_nodal_dofs; idof++ )
      dofs[idof]->add_increment_to_total();
}

void IncNode::reset_increments()
{
   for ( int idof=0; idof<nr_nodal_dofs; idof++ )
      dofs[idof]->set_increment( 0.0 );
}

void IncNode::add_to_increments( const Vector &free, const Vector &prescribed )
{
   for ( int idof=0; idof<nr_nodal_dofs; idof++ )
   {
      if ( dofs[idof]->is_free() )
         dofs[idof]->add_to_increment( free( dofs[idof]->get_global_dof() ) );
      else if ( dofs[idof]->is_prescribed() )
         dofs[idof]->add_to_increment( prescribed( dofs[idof]->get_global_dof() ) );
   }
}

void IncNode::add_to_increments( const Vector &free )
{
   for ( int idof=0; idof<nr_nodal_dofs; idof++ )
   {
      if ( dofs[idof]->is_free() )
         dofs[idof]->add_to_increment( free( dofs[idof]->get_global_dof() ) );
   }
}

void IncNode::get_incremental_solution_vector( Vector &free ) const
{
   for ( int idof=0; idof<nr_nodal_dofs; idof++ )
   {
      if ( dofs[idof]->is_free() )
         free( dofs[idof]->get_global_dof() ) = dofs[idof]->get_increment();
   }
}

CartVector IncNode::get_incremental_displacement() const
{
   double     value;
   CartVector U(0,0,0);
   Direction  dir;

   for ( int idof=0; idof<nr_nodal_dofs; idof++ )
   {
      if ( dofs[idof]->get_dof_type() == Dof::u )
      {
         value = dofs[idof]->get_increment();

         dir = dofs[idof]->get_direction();
         U += dir*value;
      }
   }

   return U;
}

double IncNode::get_temperature_increment() const
{
   for ( int idof=0; idof<nr_nodal_dofs; idof++ )
   {
      if ( dofs[idof]->get_dof_type() == Dof::t )
         return dofs[idof]->get_increment();
   }

   return 0.0;
}
