// Copyright (c) 1999, A.H. van den Boogaard

#include "shape.h"
#include "ftexcept.h"

using FeaTure::Vector;
using FeaTure::Matrix;
using FeaTure::IntLocs;
using FeaTure::Gauss;
using FeaTure::Hammer;
using FeaTure::Triangular;
using FeaTure::Tri_lin;
using FeaTure::Tri_quad;
using FeaTure::Tri_quad_alt;
using FeaTure::Quadrilateral;
using FeaTure::Quad_lin;
using FeaTure::Quad_quad;
using FeaTure::Quad_quad_alt;
using FeaTure::Hexagonal;
using FeaTure::Hex_lin;

IntLocs Triangular::get_int_locs() const
{
   double w;
   Vector Xi(3);

   IntLocs intlocs(int_scheme);

   for ( int ip=1; ip<=get_nr_int_points(); ip++ )
   {
      Hammer( ip, int_scheme, Xi[0], Xi[1], Xi[2], w );
      intlocs.set_values( ip, Xi, w );
   }

   return intlocs;
}

IntLocs Triangular::get_int_locs_face( int face_nr ) const
{
   double xi, L, w;
   int    nr_line_ip, i, j, k;
   Vector Xi(3);

   switch ( face_nr )
   {  /* L(i) 0->1, L(j) 1->0, L(k)=0 */
      case 1:  i=2; j=1; k=3;
               break;
      case 2:  i=3; j=2; k=1;
               break;
      case 3:  i=1; j=3; k=2;
               break;
      default: throw program_error( "illegal face number in Triangular::get_int_locs_face()" );
   }

   if ( int_scheme <= 3 )
      nr_line_ip = 2;
   else if ( int_scheme <= 7 )
      nr_line_ip = 3;
   else
      nr_line_ip = 4;

   IntLocs intlocs(nr_line_ip);

   for ( int ip=1; ip<=nr_line_ip; ip++ )
   {
      Gauss( ip, nr_line_ip, xi, w );
      L = 0.5*(xi+1);
      Xi(i) = L;
      Xi(j) = 1.0-L;
      Xi(k) = 0.0;
      intlocs.set_values( ip, Xi, w );
   }

   return intlocs;
}

Matrix Triangular::dVdA( int face_nr ) const
{
   Matrix dvda(2,1);

   switch ( face_nr )
   {
      case 1: dvda(1,1) = -0.5;
              dvda(2,1) =  0.5;
              break;
      case 2: dvda(1,1) =  0.0;
              dvda(2,1) = -0.5;
              break;
      case 3: dvda(1,1) =  0.5;
              dvda(2,1) =  0.0;
              break;
      default:throw program_error( "illegal face number in Triangular::dVda()" );
   }

   return dvda;
}

Matrix Tri_lin::get_N_vec( const Vector &Xi ) const
{
   double L1 = Xi[0], L2 = Xi[1], L3 = Xi[2];

   Matrix N(1,3);

   N(1,1) = L1;
   N(1,2) = L2;
   N(1,3) = L3;

   return N;
}

Matrix Tri_lin::getPartialDerivatives( const Vector& /*natcor*/ ) const
{
   Matrix P(2,3);

   P(1,1) =  1.0; P(2,1) =  0.0;
   P(1,2) =  0.0; P(2,2) =  1.0;
   P(1,3) = -1.0; P(2,3) = -1.0;

   return P;
}

Matrix Tri_quad::get_N_vec( const Vector &Xi ) const
{
   double L1 = Xi[0], L2 = Xi[1], L3 = Xi[2];

   Matrix N(1,6);

   N(1,1) = L1*(2*L1 - 1);
   N(1,2) = 4*L1*L2;
   N(1,3) = L2*(2*L2 - 1);
   N(1,4) = 4*L2*L3;
   N(1,5) = L3*(2*L3 - 1);
   N(1,6) = 4*L3*L1;

   return N;
}

Matrix Tri_quad::getPartialDerivatives( const Vector &Xi ) const
{
   double L1 = Xi[0], L2 = Xi[1];//, L3 = Xi[2];

   Matrix P(2,6);

   P(1,1) =  4*L1 - 1;        P(2,1) =  0.0;
   P(1,2) =  4*L2;            P(2,2) =  4*L1;
   P(1,3) =  0.0;             P(2,3) =  4*L2 - 1;
   P(1,4) = -4*L2;            P(2,4) =  4 - 8*L2 - 4*L1;
   P(1,5) = -3 + 4*(L1+L2);   P(2,5) = -3 + 4*(L1+L2);
   P(1,6) =  4 - 8*L1 - 4*L2; P(2,6) = -4*L1;

   return P;
}

Matrix Tri_quad_alt::get_N_vec( const Vector &Xi ) const
{
   double L1 = Xi[0], L2 = Xi[1], L3 = Xi[2];

   Matrix N(1,6);

   N(1,1) = L1*(2*L1 - 1);
   N(1,4) = 4*L1*L2;
   N(1,2) = L2*(2*L2 - 1);
   N(1,5) = 4*L2*L3;
   N(1,3) = L3*(2*L3 - 1);
   N(1,6) = 4*L3*L1;

   return N;
}

Matrix Tri_quad_alt::getPartialDerivatives( const Vector &Xi ) const
{
   double L1 = Xi[0], L2 = Xi[1];//, L3 = Xi[2];

   Matrix P(2,6);

   P(1,1) =  4*L1 - 1;        P(2,1) =  0.0;
   P(1,4) =  4*L2;            P(2,4) =  4*L1;
   P(1,2) =  0.0;             P(2,2) =  4*L2 - 1;
   P(1,5) = -4*L2;            P(2,5) =  4 - 8*L2 - 4*L1;
   P(1,3) = -3 + 4*(L1+L2);   P(2,3) = -3 + 4*(L1+L2);
   P(1,6) =  4 - 8*L1 - 4*L2; P(2,6) = -4*L1;

   return P;
}

IntLocs Quadrilateral::get_int_locs() const
{
   int xip, eip;
   double w1, w2;
   Vector natcor(2);

   IntLocs intlocs(get_nr_int_points());
   int ip = 0;

   for ( xip=1; xip<=int_scheme[0]; xip++ )
   {
      Gauss( xip, int_scheme[0], natcor[0], w1 );
      for ( eip=1; eip<= int_scheme[1]; eip++ )
      {
         Gauss( eip, int_scheme[1], natcor[1], w2 );

         intlocs.set_values( ++ip, natcor, w1*w2 );
      }
   }
   return intlocs;
}

IntLocs Quadrilateral::get_int_locs_face( int face_nr ) const
{
   double w;
   int    nr_line_ip, i;
   Vector Xi(2,0.0);

   switch ( face_nr )
   {  /* i is active local coordinate */
      case 1:  i=1; Xi(2)=-1; nr_line_ip=int_scheme[0];
               break;
      case 2:  i=2; Xi(1)=1;  nr_line_ip=int_scheme[1];
               break;
      case 3:  i=1; Xi(2)=1;  nr_line_ip=int_scheme[0];
               break;
      case 4:  i=2; Xi(1)=-1; nr_line_ip=int_scheme[1];
               break;
      default: throw program_error( "illegal face number in Quadrilateral::get_int_locs_face()" );
   }

   IntLocs intlocs(nr_line_ip);

   for ( int ip=1; ip<=nr_line_ip; ip++ )
   {
      Gauss( ip, nr_line_ip, Xi(i), w );
      intlocs.set_values( ip, Xi, w );
   }

   return intlocs;
}

Matrix Quadrilateral::dVdA( int face_nr ) const
{
   Matrix dvda(2,1);

   switch ( face_nr )
   {
      case 1: dvda(1,1) =  1.0;
              dvda(2,1) =  0.0;
              break;
      case 2: dvda(1,1) =  0.0;
              dvda(2,1) =  1.0;
              break;
      case 3: dvda(1,1) = -1.0;
              dvda(2,1) =  0.0;
              break;
      case 4: dvda(1,1) =  0.0;
              dvda(2,1) = -1.0;
              break;
      default:throw program_error( "illegal face number in Quadrilateral::dVdA()" );
   }

   return dvda;
}

Matrix Quad_lin::get_N_vec( const Vector &natcor ) const
{
   double xi = natcor(1), eta=natcor(2);

   Matrix N(1,4);

   N(1,1) = 0.25*(1-xi)*(1-eta);
   N(1,2) = 0.25*(1+xi)*(1-eta);
   N(1,3) = 0.25*(1+xi)*(1+eta);
   N(1,4) = 0.25*(1-xi)*(1+eta);

   return N;
}

Matrix Quad_lin::getPartialDerivatives( const Vector &natcor ) const
{
   double xi = natcor(1), eta=natcor(2);

   Matrix P(2,4);

   P(1,1) = -0.25*(1-eta); P(2,1) = -0.25*(1-xi);
   P(1,2) =  0.25*(1-eta); P(2,2) = -0.25*(1+xi);
   P(1,3) =  0.25*(1+eta); P(2,3) =  0.25*(1+xi);
   P(1,4) = -0.25*(1+eta); P(2,4) =  0.25*(1-xi);

   return P;
}

Matrix Quad_quad::get_N_vec( const Vector &natcor ) const
{
   double xi = natcor(1), eta=natcor(2);

   Matrix N(1,8);

   N(1,1) = -0.25*(1-xi)*(1-eta)*(1+xi+eta);
   N(1,2) = 0.5*(1-xi*xi)*(1-eta);
   N(1,3) = -0.25*(1+xi)*(1-eta)*(1-xi+eta);
   N(1,4) = 0.5*(1+xi)*(1-eta*eta);
   N(1,5) = -0.25*(1+xi)*(1+eta)*(1-xi-eta);
   N(1,6) = 0.5*(1-xi*xi)*(1+eta);
   N(1,7) = -0.25*(1-xi)*(1+eta)*(1+xi-eta);
   N(1,8) = 0.5*(1-xi)*(1-eta*eta);

   return N;
}

Matrix Quad_quad::getPartialDerivatives( const Vector &natcor ) const
{
   double xi = natcor(1), eta=natcor(2);

   Matrix P(2,8);

   P(1,1) = 0.25*(1-eta)*(2*xi+eta); P(2,1) = 0.25*(1-xi)*(2*eta+xi);
   P(1,2) = xi*(eta-1);              P(2,2) = 0.5*(xi*xi-1);
   P(1,3) = 0.25*(1-eta)*(2*xi-eta); P(2,3) = 0.25*(1+xi)*(2*eta-xi);
   P(1,4) = 0.5*(1-eta*eta);         P(2,4) = -eta*(1+xi);
   P(1,5) = 0.25*(1+eta)*(2*xi+eta); P(2,5) = 0.25*(1+xi)*(2*eta+xi);
   P(1,6) = -xi*(1+eta);             P(2,6) = 0.5*(1-xi*xi);
   P(1,7) = 0.25*(1+eta)*(2*xi-eta); P(2,7) = 0.25*(1-xi)*(2*eta-xi);
   P(1,8) = 0.5*(eta*eta-1);         P(2,8) = eta*(xi-1);
   return P;
}

Matrix Quad_quad_alt::get_N_vec( const Vector &natcor ) const
{
   double xi = natcor(1), eta=natcor(2);

   Matrix N(1,8);

   N(1,1) = -0.25*(1-xi)*(1-eta)*(1+xi+eta);
   N(1,5) = 0.5*(1-xi*xi)*(1-eta);
   N(1,2) = -0.25*(1+xi)*(1-eta)*(1-xi+eta);
   N(1,6) = 0.5*(1+xi)*(1-eta*eta);
   N(1,3) = -0.25*(1+xi)*(1+eta)*(1-xi-eta);
   N(1,7) = 0.5*(1-xi*xi)*(1+eta);
   N(1,4) = -0.25*(1-xi)*(1+eta)*(1+xi-eta);
   N(1,8) = 0.5*(1-xi)*(1-eta*eta);

   return N;
}

Matrix Quad_quad_alt::getPartialDerivatives( const Vector &natcor ) const
{
   double xi = natcor(1), eta=natcor(2);

   Matrix P(2,8);

   P(1,1) = 0.25*(1-eta)*(2*xi+eta); P(2,1) = 0.25*(1-xi)*(2*eta+xi);
   P(1,5) = xi*(eta-1);              P(2,5) = 0.5*(xi*xi-1);
   P(1,2) = 0.25*(1-eta)*(2*xi-eta); P(2,2) = 0.25*(1+xi)*(2*eta-xi);
   P(1,6) = 0.5*(1-eta*eta);         P(2,6) = -eta*(1+xi);
   P(1,3) = 0.25*(1+eta)*(2*xi+eta); P(2,3) = 0.25*(1+xi)*(2*eta+xi);
   P(1,7) = -xi*(1+eta);             P(2,7) = 0.5*(1-xi*xi);
   P(1,4) = 0.25*(1+eta)*(2*xi-eta); P(2,4) = 0.25*(1-xi)*(2*eta-xi);
   P(1,8) = 0.5*(eta*eta-1);         P(2,8) = eta*(xi-1);
   return P;
}

IntLocs Hexagonal::get_int_locs() const
{
   int xip, eip, zip;
   double w1, w2, w3;
   Vector natcor(3);

   IntLocs intlocs(get_nr_int_points());
   int ip = 0;

   for ( xip=1; xip<=int_scheme[0]; xip++ )
   {
      Gauss( xip, int_scheme[0], natcor(1), w1 );
      for ( eip=1; eip<= int_scheme[1]; eip++ )
      {
         Gauss( eip, int_scheme[1], natcor(2), w2 );
         for ( zip=1; zip<=int_scheme[2]; zip++ )
         {
            Gauss( zip, int_scheme[2], natcor(3), w3 );
            intlocs.set_values( ++ip, natcor, w1*w2*w3 );
         }
      }
   }
   return intlocs;
}

IntLocs Hexagonal::get_int_locs_face( int face_nr ) const
{
   double w1, w2;
   int    ip, nr_face_ip, i, j;
   Vector Xi(3,0);

   switch ( face_nr )
   {  /* i and j are the active local coordinates */
      case 1:  i=1; j=3; Xi(2)=-1;
               break;
      case 2:  i=1; j=3; Xi(2)=+1;
               break;
      case 3:  i=1; j=2; Xi(3)=-1;
               break;
      case 4:  i=2; j=3; Xi(1)= 1;
               break;
      case 5:  i=1; j=2; Xi(3)= 1;
               break;
      case 6:  i=2; j=3; Xi(1)=-1;
               break;
      default: throw program_error( "illegal face number in Hexagonal::get_int_locs_face()" );
   }

   nr_face_ip = int_scheme[i-1]*int_scheme[j-1];
   IntLocs intlocs(nr_face_ip);

   ip=0;
   for ( int xip=1; xip<=int_scheme[i-1]; xip++ )
   {
      Gauss( xip, int_scheme[i-1], Xi(i), w1 );
      for ( int eip=1; eip<=int_scheme[j-1]; eip++ )
      {
         Gauss( eip, int_scheme[j-1], Xi(j), w2 );
         intlocs.set_values( ++ip, Xi, w1*w2 );
      }
   }

   return intlocs;
}

Matrix Hexagonal::dVdA( int face_nr ) const
{
   Matrix dvda(3,2,0.0);

   switch ( face_nr )
   {
      case 1: dvda(1,1) =  1.0;
              dvda(3,2) =  1.0;
              break;
      case 2: dvda(1,1) = -1.0;
              dvda(3,2) =  1.0;
              break;
      case 3: dvda(1,1) = -1.0;
              dvda(2,2) =  1.0;
              break;
      case 4: dvda(2,1) =  1.0;
              dvda(3,2) =  1.0;
              break;
      case 5: dvda(1,1) =  1.0;
              dvda(2,2) =  1.0;
              break;
      case 6: dvda(2,1) = -1.0;
              dvda(3,2) =  1.0;
              break;
      default:throw program_error( "illegal face number in Hexagonal::dVdA()" );
   }

   return dvda;
}

Matrix Hex_lin::get_N_vec( const Vector &natcor ) const
{
   double xi = natcor(1), eta=natcor(2), zeta=natcor(3);

   Matrix N(1,8);

   N(1,1) = 0.125*(1-xi)*(1-eta)*(1-zeta);
   N(1,2) = 0.125*(1+xi)*(1-eta)*(1-zeta);
   N(1,3) = 0.125*(1+xi)*(1+eta)*(1-zeta);
   N(1,4) = 0.125*(1-xi)*(1+eta)*(1-zeta);
   N(1,5) = 0.125*(1-xi)*(1-eta)*(1+zeta);
   N(1,6) = 0.125*(1+xi)*(1-eta)*(1+zeta);
   N(1,7) = 0.125*(1+xi)*(1+eta)*(1+zeta);
   N(1,8) = 0.125*(1-xi)*(1+eta)*(1+zeta);

   return N;
}

Matrix Hex_lin::getPartialDerivatives( const Vector &natcor ) const
{
   double xi = natcor(1), eta=natcor(2), zeta=natcor(3);

   Matrix P(3,8);

   P(1,1) = -0.125*(1-eta)*(1-zeta); P(2,1) = -0.125*(1-xi)*(1-zeta);  P(3,1) = -0.125*(1-xi)*(1-eta);
   P(1,2) =  0.125*(1-eta)*(1-zeta); P(2,2) = -0.125*(1+xi)*(1-zeta);  P(3,2) = -0.125*(1+xi)*(1-eta);
   P(1,3) =  0.125*(1+eta)*(1-zeta); P(2,3) =  0.125*(1+xi)*(1-zeta);  P(3,3) = -0.125*(1+xi)*(1+eta);
   P(1,4) = -0.125*(1+eta)*(1-zeta); P(2,4) =  0.125*(1-xi)*(1-zeta);  P(3,4) = -0.125*(1-xi)*(1+eta);
   P(1,5) = -0.125*(1-eta)*(1+zeta); P(2,5) = -0.125*(1-xi)*(1+zeta);  P(3,5) =  0.125*(1-xi)*(1-eta);
   P(1,6) =  0.125*(1-eta)*(1+zeta); P(2,6) = -0.125*(1+xi)*(1+zeta);  P(3,6) =  0.125*(1+xi)*(1-eta);
   P(1,7) =  0.125*(1+eta)*(1+zeta); P(2,7) =  0.125*(1+xi)*(1+zeta);  P(3,7) =  0.125*(1+xi)*(1+eta);
   P(1,8) = -0.125*(1+eta)*(1+zeta); P(2,8) =  0.125*(1-xi)*(1+zeta);  P(3,8) =  0.125*(1-xi)*(1+eta);

   return P;
}
