// Copyright (c) 1999, A.H. van den Boogaard

#include "transmat.h"
#include "ftexcept.h"
#include <cassert>

using FeaTure::TransformationMatrix;
using FeaTure::Matrix;
using FeaTure::Vector;

void TransformationMatrix::put( size_t i, size_t j, double val )
{
   assert( filled < nr_nonzeros );
   assert( i>0 && j>0 && i<=dim1 && j<=dim2 );

   // every element may only be initialized once
   for ( size_t m=0; m<filled; m++ ) assert( !( a[m].i==i && a[m].j==j ) );

   a[filled].i = i;
   a[filled].j = j;
   a[filled].val = val;
   filled++;

   if ( filled == nr_nonzeros )
   {
      // evaluate for unit matrix or permutation matrix
      bool unit=(dim1==dim2), permutation=true;
      for ( size_t m=0; m<nr_nonzeros; m++ )
      {
         if ( a[m].val != 1.0 ) { unit=permutation=false; break; }
         if ( a[m].i != a[m].j ) { unit=false; }
      }

      if ( unit )
      {
         delete [] a;
         a=NULL;
         type = UNIT;
      }
      else if ( permutation )
      {
         p = new int[dim1];
         for ( size_t m=0; m<dim1; m++ ) p[m]=-1;
         for ( size_t m=0; m<nr_nonzeros; m++ )
         {
            assert( p[(a[m].i)-1]== -1 );
            p[(a[m].i)-1] = a[m].j;
         }
         delete [] a;
         a=NULL;
         type = PERMUTATION;
      }
   }
}

double TransformationMatrix::get( size_t i, size_t j ) const
{
   size_t m;

   assert( i<=dim1 && j<=dim2 );
   switch ( type )
   {
     case UNIT:   return ( i==j ? 1.0 : 0.0 );
     case PERMUTATION:
                  return ( p[i-1]==int(j) ? 1.0 : 0.0 );
     case SPARSE: for ( m=0; m<nr_nonzeros; m++ )
                     if ( a[m].i==i && a[m].j==j ) return a[m].val;
                  return 0.0;
     default:     throw program_error( "TransformationMatrix not initialized yet" );
   }
}

Vector TransformationMatrix::mult( const Vector &v ) const
{
   assert ( v.size() == dim2 );
   if ( type == UNIT ) return v;
   else if ( type == PERMUTATION )
   {
      Vector r;
      r.resize(dim1);
      for ( size_t m=0; m<dim1; m++ ) r[m] = (p[m]==-1 ? 0 : v[p[m]-1]);
      return r;
   }
   else if ( type == SPARSE )
   {
     Vector r(dim1,0.0);
     for ( size_t m=0; m<nr_nonzeros; m++ ) r(a[m].i) += a[m].val*v(a[m].j);
     return r;
   }
   else
      throw program_error( "TransformationMatrix not initialized yet" );
}

Vector TransformationMatrix::tmult( const Vector &v ) const
{
   assert( v.size() == dim1 );

   if ( type == UNIT ) return v;
   else if ( type == PERMUTATION )
   {
      Vector r(dim2,0.0);
      for ( size_t m=0; m<dim1; m++ )
         if ( p[m]!=-1 ) r[p[m]-1] += v[m];
      return r;
   }
   else if ( type == SPARSE )
   {
      Vector r(dim2,0.0);
      for ( size_t m=0; m<nr_nonzeros; m++ )
         r(a[m].j) += a[m].val*v(a[m].i);
      return r;
   }
   else
      throw program_error( "TransformationMatrix not initialized yet" );
}

Matrix TransformationMatrix::preTmult_postMult( const Matrix &M ) const
{
   assert( M.size1() == dim1 && M.size2() == dim1 );

   if ( type == UNIT ) return M;
   else if ( type == PERMUTATION )
   {
      Matrix R(dim2,dim2,0.0);
      for ( size_t m=0; m<dim1; m++ )
      {
         if ( p[m]==-1 ) continue;
         for ( size_t k=0; k<dim1; k++ )
            if ( p[k]!=-1 ) R(p[m],p[k]) += M[m][k];
      }
      return R;
   }
   else if ( type == SPARSE )
   {
      Matrix R(dim2,dim2,0.0);
      Matrix MT(dim1,dim2,0.0);
      /* R = T^t*M*T , first MT = M*T, then R=T^t*MT */
      for ( size_t m=0; m<nr_nonzeros; m++ )
      {
         size_t row = a[m].i-1;
         size_t col = a[m].j-1;
         double val = a[m].val;

         for ( size_t i=0; i<dim1; i++ )
            MT[i][col] += val*M[i][row];
      }
           // now R=T^t*MT
      for ( size_t m=0; m<nr_nonzeros; m++ )
      {
         size_t row = a[m].j-1;
         size_t col = a[m].i-1;
         double val = a[m].val;

         for ( size_t j=0; j<dim2; j++ )
            R[row][j] += val*MT[col][j];
      }
      return R;
   }
   else
      throw program_error( "TransformationMatrix not initialized yet" );
}

std::ostream& operator<<(std::ostream &s, const TransformationMatrix &A)
{
    size_t M=A.num_rows();
    size_t N=A.num_cols();

    s << M << "x" << N << std::endl;

    for (size_t i=1; i<=M; i++)
    {
        for (size_t j=1; j<=N; j++)
        {
            s << A.get(i,j) << " ";
        }
        s << std::endl;
    }

    return s;
}
